<?php

namespace App\Console;

use App\Services\DelayedActions;
use Illuminate\Console\Scheduling\Schedule;
use Laravel\Lumen\Console\Kernel as ConsoleKernel;

class Kernel extends ConsoleKernel
{
    /**
     * The Artisan commands provided by your application.
     *
     * @var array
     */
    protected $commands = [
        Commands\KeycloakSync::class
    ];

    /**
     * Define the application's command schedule.
     *
     * @param  \Illuminate\Console\Scheduling\Schedule  $schedule
     * @return void
     */
    protected function schedule(Schedule $schedule)
    {
        // First scheduled task creates the jobs from the delayed actions table,
        // if they have reached the action date.
        $schedule->call(new DelayedActions)
            ->name('delayedActions')
            ->withoutOverlapping();

        // Second scheduled task executes the jobs created in the previous step.
        $schedule->command('queue:work --queue=updowngrade,cancel,change,delete --stop-when-empty')
            ->dailyAt('2:00')
            ->runInBackground();

        // Third scheduled task has nothing to do with the previous tasks,
        // but executes the daily syncing with the Keycloak server.
        // Daily currently disabled because all of its functionality is done by the NOW sync.
        //        $schedule->command('keycloak:sync daily')
        //            ->dailyAt('3:00')
        //            ->runInBackground();

        // Fourth scheduled task executes the hourly syncing with the Keycloak server.
        $schedule->command('keycloak:sync hourly')
            ->hourly()
            ->name('keycloakSyncHourly')
            ->withoutOverlapping()
            ->runInBackground();

        // Check for pending syncs every 5 minutes
        $schedule->command('queue:work --queue=sync --stop-when-empty')
            ->everyFiveMinutes()
            ->withoutOverlapping()
            ->runInBackground();
    }
}
