<?php
/**
 * verein.cloud SPI API
 * ## Info ## API that allows interfacing of a Keycloak User Storage SPI with the Benutzer table of S-Verein.  Make sure to always view the model and not just the examples, because the models contain a lot of important descriptions for the fields that are otherwise not visible in Swagger.  See the GIT readme.md file for further information.
 * PHP version 7.2.5
 *
 * The version of the OpenAPI document: 1.0.0
 * Contact: mheumann@comciencia.cl
 *
 * NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).
 * https://openapi-generator.tech
 * Do not edit the class manually.
 */

namespace App\Http\Controllers;

use App\Services\DirectAccess;
use App\Models\Login;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Request;
use App\Services\VVInterfaceBB;
use App\Models\Databases;

class BBApi extends Controller
{
    /**
     * Constructor
     */
    public function __construct()
    {
    }

    /**
     * Operation usersGet
     * @return \Illuminate\Http\JsonResponse response
     */
    public function usersGet()
    {
        $input = Request::all();

        if (!isset($input['vid'])) {
            throw new \InvalidArgumentException('Missing the required parameter $vid when calling usersIdGet');
        }
        if (strlen($input['vid']) > 50) {
            throw new \InvalidArgumentException('invalid length for $vid when calling BBApi.usersIdGet, must be smaller than or equal to 50.');
        }
        $vid = $input['vid'];

        $this->validate(Request::instance(), [
            'first' => ['required', 'integer', 'min:0', 'max:99999999'],
            'max' => ['required', 'integer', 'min:0', 'max:99999999'],
            'search' => ['required', 'string', 'min:1', 'max:100', 'regex:/^[\d\s,.\'\-\pL]+$/u'],
        ]);

        $first = $input['first'];
        $max = $input['max'];
        $search = $input['search'];

        $aParts = explode( "_", $vid );

        $database_id = $aParts[0];
        $club_id = $aParts[1];

        if( empty($database_id) || empty($club_id) ) {
            abort(400, 'Invalid vid');
        }

        // Initialize VV environment
        $vvEnv = new VVInterfaceBB($database_id, $club_id);

        // Call VV function
        $aData = $vvEnv->vvFindUser($club_id, $first, $max, $search);

        if ( !empty($aData)) {
            return response()->json( $aData );
        }
        else {
            abort(404, 'User not found');
        }
    }

    /**
     * Operation usersCountGet
     * @return \Illuminate\Http\JsonResponse response
     */
    public function usersCountGet()
    {
        $input = Request::all();

        if (!isset($input['vid'])) {
            throw new \InvalidArgumentException('Missing the required parameter $vid when calling usersCountGet');
        }
        if (strlen($input['vid']) > 50) {
            throw new \InvalidArgumentException('invalid length for $vid when calling BBApi.usersCountGet, must be smaller than or equal to 50.');
        }
        $vid = $input['vid'];

        $aParts = explode( "_", $vid );

        $database_id = $aParts[0];
        $club_id = $aParts[1];

        if( empty($database_id) || empty($club_id) ) {
            abort(400, 'Invalid vid');
        }

        // Initialize VV environment
        $vvEnv = new VVInterfaceBB($database_id, $club_id);

        // Call VV function
        $iCount = $vvEnv->vvGetUserCount($club_id);

        if ( !empty($iCount)) {
            return response()->json( $iCount );
        }
        else {
            abort(404, 'No users found');
        }
    }

    /**
     * Operation usersIdGet
     *
     * @param string $id  (required)
     *
     * @return \Illuminate\Http\JsonResponse response
     */
    public function usersIdGet($id)
    {
        if ( !isset($id) ) {
            throw new \InvalidArgumentException('Missing the required parameter $id when calling usersIdGet');
        }

        if( preg_match("/^[0-9_]+$/", $id ) == 1 ) {
            // Id is numeric
            $aParts = explode( "_", $id );
            if( count($aParts) != 3 ) {
                abort(400, 'Invalid id');
            }
            $database_id = $aParts[0];

            // Get database data from database_id
            $dbdata = Databases::getIfexists($database_id);
            if( $dbdata === FALSE ) {
                abort( 400, 'Invalid database id' );
            }

            $club_id = $aParts[1];
            $user_id = $aParts[2];

            $vvClient = new DirectAccess($dbdata);
            $shortkey = $vvClient->vvGetShortkey($club_id);

            $username = "";
            $vvClient->vvClose();
        }
        else {
            // Id should be user@shortkey
            $aParts = explode( "@", $id );

            $username = $aParts[0];
            $shortkey = $aParts[1];
            $user_id = "";

            if( empty($username) || empty($shortkey) ) {
                abort(400, 'Invalid username');
            }

            // Get database id from shortkey
            $loginDb = new Login();
            $database_id = $loginDb->getDatabaseId($shortkey);

            if( empty($database_id) ) {
                abort( 405, "No database connection possible username $username, shortkey $shortkey" );
            }

            // Get database data from database_id
            $dbdata = Databases::getIfexists($database_id);
            if( $dbdata === FALSE ) {
                abort( 400, 'Invalid database id' );
            }

            $vvDbClient = new DirectAccess($dbdata);
            $club_id = $vvDbClient->vvGetKundeId($shortkey);

            if( empty($club_id) ) {
                abort(400, 'Invalid club_id');
            }

            $vvDbClient->vvClose();
        }

        // Initialize VV environment
        $vvEnv = new VVInterfaceBB($database_id, $club_id);

        // Call VV function
        if( empty($user_id) && !empty($username) ) {
            // Standard case: user@club
            $aData = $vvEnv->vvGetUser($database_id, $club_id, $username, null, $shortkey);
        }
        else {
            $aData = $vvEnv->vvGetUser($database_id, $club_id, null, $user_id, $shortkey);
        }

        if ( !empty($aData)) {
            return response()->json( $aData );
        }
        else {
            abort(404, 'User not found');
        }
    }

    /**
     * Operation validateUser
     *
     * Validates a user through the Benutzer table of S-Verein..
     *
     * @return \Illuminate\Http\JsonResponse response
     */
    public function validateUser()
    {
        $input = Request::all();

        if (empty($input['user']) ) {
            throw new \InvalidArgumentException('Missing the required parameter $user when calling validateUser');
        }
        if (strlen($input['user']) > 50) {
            throw new \InvalidArgumentException('invalid length for $user when calling BBApi.validateUser, must be smaller than or equal to 50.');
        }
        $user = $input['user'];

        if (!isset($input['password_hash'])) {
            throw new \InvalidArgumentException('Missing the required parameter $password_hash when calling validateUser');
        }
        if (strlen($input['password_hash']) > 100) {
            throw new \InvalidArgumentException('invalid length for $password_hash when calling BBApi.validateUser, must be smaller than or equal to 100.');
        }
        $password_hash = $input['password_hash'];

        if( preg_match("/^[0-9_]+$/", $user ) == 1 ) {
            // Id is numeric
            $aParts = explode( "_", $user );
            if( count($aParts) != 3 ) {
                abort(400, 'Invalid id');
            }
            $database_id = $aParts[0];

            $club_id = $aParts[1];
            $user_id = $aParts[2];
        }
        else {
            throw new \InvalidArgumentException('Invalid user format.');
        }

        if( empty($club_id) ) {
            abort(400, 'Invalid club_id');
        }

        // Initialize VV environment
        $vvEnv = new VVInterfaceBB($database_id, $club_id);

        // Call VV function
        $iRetCode = $vvEnv->vvValidateUser($club_id, $user_id, $password_hash);

        if ($iRetCode == 200) {
            Log::debug( "Validation successful." );
            return response("OK", 200);
        }
        else {
            Log::debug( "Validation failed, returncode $iRetCode." );
            switch ( $iRetCode ) {
                case 401:
                    abort(401, 'Unauthorized access');
                    break;
                case 403:
                    abort(403, 'Access forbidden, no permissions');
                    break;
                case 404:
                    abort(404, 'Invalid username or password');
                    break;
                default:
                    abort(500, 'Unexpected error');
            }
        }
    }

    /**
     * Operation usersGet
     * @return \Illuminate\Http\JsonResponse response
     */
    public function deliverClientCreds()
    {
        $input = Request::all();

        $this->validate(Request::instance(), [
            'tenant' => ['required', 'string', 'min:1', 'max:50', 'regex:/^[\w\d]+$/iu'],
            'customers_id' => ['required', 'numeric', 'min:1', 'max:999999999999'],
            'users_id' => ['sometimes', 'numeric', 'min:1', 'max:999999999999'],
            'username' => ['required_with:users_id', 'string', 'min:1', 'max:50'],
            'api_client' => ['required', 'string', 'min:1', 'max:100'],
            'api_secret' => ['required', 'string', 'min:1', 'max:255'],
            'api_key' => ['required', 'string', 'min:1', 'max:255'],
        ]);

        $vid = $input['tenant'];
        $bbCustId = $input['customers_id'];
        $apiClient = $input['api_client'];
        $apiSecret = $input['api_secret'];
        $apiKey = $input['api_key'];
        if( isset($input['users_id']) ) {
            $bbUserId = $input['users_id'];
        }
        else {
            $bbUserId = "";
        }
        if( isset($input['username']) ) {
            $vvUsername = $input['username'];
        }
        else {
            $vvUsername = "";
        }

        $loginDb = new Login();
        $database_id = $loginDb->getDatabaseId($vid);

        if( empty($database_id) ) {
            Log::debug( "deliverClientCreds abort 400" );
            abort(400, 'Invalid vid');
        }

        // Get database data from database_id
        $dbdata = Databases::getIfexists($database_id);
        if( $dbdata === FALSE ) {
            abort( 400, 'Invalid database id' );
        }

        $vvClient = new DirectAccess($dbdata);
        $club_id = $vvClient->vvGetKundeId($vid);
        $vvClient->vvClose();

        // Initialize VV environment
        $vvEnv = new VVInterfaceBB($database_id, $club_id);

        $aCredData = array(
            "bbcustid" => $bbCustId,
            "apiclient" => $apiClient,
            "apisecret" => $apiSecret,
            "apikey" => $apiKey,
            "bbuserid" => $bbUserId,
            "vvusername" => $vvUsername
        );

        // Call VV function
        $bSuccess = $vvEnv->vvSaveBBCreds($club_id, $aCredData);

        if ( $bSuccess ) {
            Log::debug( "deliverClientCreds success 200" );
            return response("OK", 200);
        }
        else {
            Log::debug( "deliverClientCreds abort 500" );
            abort(500, "Couldn't store data");
        }
    }
}
