<?php

/**
 * verein.cloud CRM API
 * ## Info ## API that allows creation, modification and deletion of clubs for S-Verein. This API is meant for use by a Tineon CRM system.  See the GIT readme.md file for further information.
 * PHP version 7.2.5
 *
 * The version of the OpenAPI document: 1.0.3
 * Contact: mheumann@comciencia.cl
 *
 * NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).
 * https://openapi-generator.tech
 * Do not edit the class manually.
 */


namespace App\Http\Controllers;

use App\Models\Delayed;
use App\Services\DelayedActions;
use App\Services\DirectAccess;
use App\Services\VVInterfaceCRM;
use App\Models\Databases;
use App\Models\Login;
use App\Rules\ValidArrayVal;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Request;

class CRMApiController extends Controller
{
    const VALID_ATTRIBUTES = [
        "fibuLight",
        "fibuPremium",
        "small",
        "app",
        "appLight",
    ];

    // Regex for standard string validation, such as names, city, street
    // Allowed characters are numbers, any language letter, blank and ^!"§$%&/()=?`'*+#-_.:,;<>—[]{}|@€\~º
    const STD_STRING_REG_EXP = "^[\d\pL \^\!\"\$\/\(\)\?\'\*\+\-\[\]\{\}\|\\`—§%&#=_.:,;<>@€~º„“]+$";

    /**
     * Injected VVInterface for communication with the VV database
     * @var VVInterface
     */
    protected $vvEnv;

    /**
     * Constructor
     */
    public function __construct( VVInterfaceCRM $vvEnv)
    {
        $this->vvEnv = $vvEnv;
    }

    /**
     * Operation getClubStats
     *
     * Delivers statistics for the specified club..
     *
     *
     * @return Http response
     */
    public function getClubStats()
    {
        $aData = $this->vvEnv->vvGetClubStatistics( $this->vvEnv->getClubId() );

        if( $aData === FALSE ) {
            abort(500, 'Database error' );
        }
        elseif( empty($aData) ) {
            abort(404, 'No data found' );
        }

        return response()->json( $aData );
    }

    /**
     * Operation createClub
     *
     * Creates a new club in the specified database..
     *
     *
     * @return Http response
     */
    public function createClub()
    {
        $input = Request::all();

        $aCountries = $this->vvEnv->vvGetCountryList(TRUE);
        $aCountryCodes = array_column($aCountries, "value", "code");
        $aCountryIds = array_column($aCountries, "id", "code");
        $aDesigns = $this->vvEnv->vvGetDesigns( TRUE );

        // Validate input data
        $this->validate(Request::instance(), [
            'shortKey' => ['sometimes', 'string', 'min:1', 'max:50', 'regex:/^[\w\d]+$/iu'],
            'clubName' => ['required', 'string', 'min:1', 'max:255', 'regex:/'. self::STD_STRING_REG_EXP .'/u'],
            'street' => ['sometimes', 'string', 'min:1', 'max:100', 'regex:/'. self::STD_STRING_REG_EXP .'/u'],
            'postalCode' => ['sometimes', 'string', 'min:1', 'max:15', 'regex:/^[a-z0-9\- ]+$/iu'],
            'city' => ['sometimes', 'string', 'min:1', 'max:50', 'regex:/'. self::STD_STRING_REG_EXP .'/u'],
            'country' => ['sometimes', 'string', 'min:2', 'max:2', new ValidArrayVal($aCountryCodes)],
            'contactLastname' => ['sometimes', 'string', 'min:1', 'max:50', 'regex:/'. self::STD_STRING_REG_EXP .'/u'],
            'contactFirstname' => ['sometimes', 'string', 'min:1', 'max:50', 'regex:/'. self::STD_STRING_REG_EXP .'/u'],
            'email' => 'required|email|max:255',
            'numberOfUsers' => ['required', 'integer', 'min:1', 'max:50'],
            'agbCheck' => ['sometimes', 'date_format:Y-m-d'],
            'design' => ['required', 'integer', 'min:1', new ValidArrayVal($aDesigns)],
            'testMode' => ['required', 'boolean'],
            'testEndDate' => ['exclude_unless:testMode,true', 'required', 'date_format:Y-m-d'],
        ]);

        // Validate attributes
        if( !empty($input["attributes"]) ) {
            foreach ($input["attributes"] as $key => $value) {
                if (!in_array($key, self::VALID_ATTRIBUTES)) {
                    abort(422, 'Invalid attribute value: ' . $key);
                }
                if( (is_string($value) && !in_array(strtolower($value), array("true", "false"))) ||
                    (is_bool($value) && !in_array($value, array(true, false))) )    {
                    abort(422, "Attribute value $key must be true or false.");
                }
            }
        }

        if( isset($input["country"]) ) {
            $input["country"] = $aCountryIds[$input["country"]];
        }

        try {
            $login = new Login;
            if( empty($input["shortKey"]) || !$login->isShortkeyUnique($input["shortKey"]) ) {
               $strShortkey = $login->suggestShortkey( $input["clubName"] );
            }
            else {
               $strShortkey = strtolower($input["shortKey"]);
            }

            // Create new club with all its data
            $aData = $this->vvEnv->vvCreateClub($input, $this->vvEnv->getDatabaseId(), $strShortkey);

            // Create new record in CAdmin Login table
            $login->insertRecord( $this->vvEnv->getDatabaseId(), $aData->shortKey, 1 );
            $aData->adminUser = $this->vvEnv->vvCreateFirstUser($input, $aData->clubId);
            $aData->loginURL = env( "VV_URL" ) . "/login/" . $strShortkey;
        }
        catch (\Exception $exception) {
            Log::error( "Abort Msg " . $exception->getMessage());
            abort(500, $exception->getMessage());
        }

        if( $aData === FALSE ) {
            abort(500, 'Database error' );
        }
        elseif( empty($aData) ) {
            abort(404, 'No data found' );
        }

        return response()->json( $aData );
    }

    /**
     * Operation deleteClub
     *
     * Deletes the specified club from the specified database..
     *
     *
     * @return Http response
     */
    public function deleteClub()
    {
        $input = Request::all();

        // Validate input data
        $this->validate(Request::instance(), [
            'delete_date' => ['sometimes', 'nullable', 'date_format:Y-m-d'],
        ]);

        if( isset($input['delete_date']) ) {
            $delete_date = $input['delete_date'];
        }
        else {
            $delete_date = null;
        }

        // Validate ic club exists in database
        if( !$this->vvEnv->vvClubExists($this->vvEnv->getClubId()) ) {
            abort( 404, 'Specified club not found' );
        }

        $bSuccess = $this->vvEnv->vvDeleteClub( $this->vvEnv->getClubId(), $this->vvEnv->getDatabaseId(), $delete_date );

        if( !$bSuccess ) {
            abort( 500, 'Could not delete club' );
        }

        return response("OK", 200);
    }

    /**
     * Operation getGeneralStats
     *
     * Delivers statistics on all clubs in the system..
     *
     *
     * @return Http response
     */
    public function getGeneralStats()
    {
        $dbdata = Databases::all();

        if( empty($dbdata) ) {
            abort( 404, 'No databases found' );
        }

        $aData["perDatabase"] = array();

        $aPerDB = array();
        $iTotalClubs = $iActiveClubs = $iInactiveClubs = $iTrialClubs = 0;
        foreach( $dbdata as $aDbInfo ) {
            $aDbInfo = $aDbInfo->toArray();

            $aTemp = array();
            $aTemp["databaseId"] = $aDbInfo["databaseid"];
            $aTemp["dbName"] = $aDbInfo["dbname"];
            $aTemp["dbDesc"] = $aDbInfo["description"];
            $aTemp["dbType"] = $aDbInfo["dbtype"];
            $aTemp["dbActive"] = ($aDbInfo["active"] == 1 ? true : false);
            if( !$aDbInfo["active"] ) {
                $aTemp["dbInactivityReason"] = $aDbInfo["activemessage"];
            }
            else {
                $aTemp["dbInactivityReason"] = "";
            }

            try {
                if( $aDbInfo["dbtype"] == "oracle" ) {
                    $aSysInfo = $aDbInfo;
                    $aSysInfo["dbuser"] = "SYSTEM";
                    $aSysInfo["dbuserpassword"] = env("ORACLE_SYSTEM_PASS");
                    $vvSysClient = new DirectAccess($aSysInfo);
                    $aTemp["dbSize"] = $vvSysClient->vvGetDBSize();
                    unset( $vvSysClient );
                }
                else {
                    $aTemp["dbSize"] = 0;
                }

                $vvClient = new DirectAccess($aDbInfo);
                $aTemp["dbTotalClubs"] = $vvClient->vvGetTotalClubs();
                $aTemp["dbTestClubs"] = $vvClient->vvGetTrialClubs();
                $aTemp["dbActiveClubs"] = $vvClient->vvGetActiveClubs();
                $aTemp["dbInactiveClubs"] = $vvClient->vvGetInactiveClubs();

                $iTotalClubs += $aTemp["dbTotalClubs"];
                $iActiveClubs += $aTemp["dbActiveClubs"];
                $iInactiveClubs += $aTemp["dbInactiveClubs"];
                $iTrialClubs += $aTemp["dbTestClubs"];

                unset( $vvClient );
            }
            catch (\Exception $e) {
                abort( 500, 'Error getting stats for DB ' . $aDbInfo["databaseid"] . ": " . $e->getMessage() );
            }

            $aPerDB[] = $aTemp;
        }

        $aData["perDatabase"] = $aPerDB;
        $aData["totals"] = array(
             "totalClubs" => $iTotalClubs,
             "activeClubs" => $iActiveClubs,
             "inactiveClubs" => $iInactiveClubs,
             "testClubs" => $iTrialClubs
        );

        return response()->json( $aData );
    }

    /**
     * Operation getListOfDatabases
     *
     * Return a list of all available databases..
     *
     *
     * @return Http response
     */
    public function getListOfDatabases()
    {
        // Get database data from database_id
        $dbdata = Databases::all(['databaseid as id', 'description as name', 'dbtype as type', 'active']);

        if( empty($dbdata) ) {
            abort( 404, 'No databases found' );
        }

        return response()->json( $dbdata );
    }

    /**
     * Operation getListOfDesigns
     *
     * Return a list of all available designs for the specified database..
     *
     *
     * @return Http response
     */
    public function getListOfDesigns()
    {
        $aData = $this->vvEnv->vvGetDesigns();

        if( $aData === FALSE ) {
            abort(500, 'Database error' );
        }
        elseif( empty($aData) ) {
            abort(404, 'No data found' );
        }

        return response()->json( $aData );
    }

    /**
     * Operation getLoginInfo
     *
     * Delivers login info for the specified club..
     *
     *
     * @return Http response
     */
    public function getLoginInfo()
    {
        $input = Request::all();

        if( empty($input['club_id']) && empty($input["shortkey"]) ) {
            abort( 400, 'Missing club id and shortkey' );
        }

        if( !empty($input['club_id']) ) {
            $club_id = $input['club_id'];
        }
        else {
            $club_id = $this->vvEnv->getMIOIdByShortkey($input["shortkey"]);
            if( $club_id === FALSE ) {
                abort( 400, 'Invalid shortkey' );
            }
        }

        $aData = $this->vvEnv->vvGetLoginInfo($club_id);

        if( $aData === FALSE ) {
            abort(500, 'Database error' );
        }
        elseif( empty($aData) ) {
            abort(404, 'No data found' );
        }

        // Send club_id as well
        $aData["clubId"] = $club_id;

        return response()->json( $aData );
    }

    /**
     * Operation updateClub
     *
     * Updates data of an existing club in the specified database..
     *
     *
     * @return Http response
     */
    public function updateClub()
    {
        $input = Request::all();

        if( count($input) == 2 ) {
            // Only database_id and club_id are given, so we can skip the rest
            abort( 400, 'No parameters given in body' );
        }

        $aCountries = $this->vvEnv->vvGetCountryList(TRUE);
        $aCountryCodes = array_column($aCountries, "value", "code");
        $aCountryIds = array_column($aCountries, "id", "code");
        $aDesigns = $this->vvEnv->vvGetDesigns( TRUE );

        // Validate input data
        $this->validate(Request::instance(), [
            'clubName' => ['sometimes', 'string', 'min:1', 'max:255', 'regex:/'. self::STD_STRING_REG_EXP .'/u'],
            'street' => ['sometimes', 'string', 'min:1', 'max:100', 'regex:/'. self::STD_STRING_REG_EXP .'/u'],
            'postalCode' => ['sometimes', 'string', 'min:1', 'max:15', 'regex:/^[a-z0-9\- ]+$/iu'],
            'city' => ['sometimes', 'string', 'min:1', 'max:50', 'regex:/'. self::STD_STRING_REG_EXP .'/u'],
            'country' => ['sometimes', 'string', 'min:2', 'max:2', new ValidArrayVal($aCountryCodes)],
            'contactLastname' => ['sometimes', 'string', 'min:1', 'max:50', 'regex:/'. self::STD_STRING_REG_EXP .'/u'],
            'contactFirstname' => ['sometimes', 'string', 'min:1', 'max:50', 'regex:/'. self::STD_STRING_REG_EXP .'/u'],
            'email' => 'sometimes|email|max:255',
            'numberOfUsers' => ['sometimes', 'integer', 'min:1', 'max:50'],
            'agbCheck' => ['sometimes', 'date_format:Y-m-d'],
            'design' => ['sometimes', 'integer', 'min:1', new ValidArrayVal($aDesigns)],
            'testMode' => ['sometimes', 'boolean'],
            'testEndDate' => ['exclude_unless:testMode,true', 'required', 'nullable', 'date_format:Y-m-d'],
            'active' => ['required_with:changeStateDate', 'sometimes', 'boolean'],
            'changeStateDate' => ['sometimes', 'nullable', 'date_format:Y-m-d'],
            'updowngradeDate' => ['sometimes', 'nullable', 'date_format:Y-m-d'],
            'cancelDate' => ['sometimes', 'nullable', 'date_format:Y-m-d'],
            'testEndDateFIBU' => ['sometimes', 'nullable', 'date_format:Y-m-d'],
        ]);

        // Validate attributes
        if( !empty($input["attributes"]) ) {
            foreach ($input["attributes"] as $key => $value) {
                if (!in_array($key, self::VALID_ATTRIBUTES)) {
                    abort(422, 'Invalid attribute value: ' . $key);
                }
                if( (is_string($value) && !in_array(strtolower($value), array("true", "false"))) ||
                    (is_bool($value) && !in_array($value, array(true, false))) )    {
                    abort(422, "Attribute value $key must be true or false.");
                }
            }
        }

        // updowngradeDate, cancelDate and changeStateDate are processed inside vvUpdateClub

        if( isset($input["country"]) ) {
            $input["country"] = $aCountryIds[$input["country"]];
        }

        try {
            $aData = $this->vvEnv->vvUpdateClub($input, $this->vvEnv->getDatabaseId(), $this->vvEnv->getClubId());
        }
        catch (\Exception $exception) {
            Log::debug( "Abort Msg " . $exception->getMessage(), (array)$exception);
            abort(500, $exception->getMessage());
        }

        if( $aData === FALSE ) {
            abort(500, 'Database error' );
        }
        elseif( empty($aData) ) {
            abort(404, 'No data found' );
        }

        return response("OK", 200);
    }

    /**
     * Operation supportUser
     *
     * Creates or activates the support user for the given club.
     *
     *
     * @return Http response
     */
    public function supportUser()
    {
        $input = Request::all();

        if( empty($input['club_id']) && empty($input["shortkey"]) ) {
            abort( 400, 'Missing club id and shortkey' );
        }

        // Validation
        $this->validate(Request::instance(), [
            'subject' => ['sometimes', 'string', 'min:1', 'max:100', 'regex:/^[\d ,.\'\-\pL]+$/u'],
        ]);

        if( !empty($input['club_id']) ) {
            $club_id = $input['club_id'];
        }
        else {
            $club_id = $this->vvEnv->getMIOIdByShortkey($input["shortkey"]);
            if( $club_id === FALSE ) {
                abort( 400, 'Invalid shortkey' );
            }
        }

        try {
            $aData = $this->vvEnv->vvActivateSupportUser($club_id, $input["subject"]);
        }
        catch( \Exception $e ) {
            Log::debug( "Error in supportUser (".__LINE__."), Abort Msg " . $e->getMessage(), (array)$e );
            abort( 500, $e->getMessage() );
        }

        if( $aData === FALSE ) {
            Log::debug( "Error in supportUser (".__LINE__."), Abort Msg vvActivateSupportUser returned FALSE" );
            abort(500, 'Database error' );
        }
        elseif( empty($aData) ) {
            abort(404, 'No data found' );
        }

        return response()->json( $aData );
    }

   /**
     * Operation getClubInfo
     *
     * Delivers information about the specified club.
     *
     * @return Http response
     */
    public function getClubInfo()
    {
        $input = Request::all();

        if( empty($input['club_id']) && empty($input["shortkey"]) ) {
            abort( 400, 'Missing club id and shortkey' );
        }

        if( !empty($input['club_id']) ) {
            $club_id = $input['club_id'];
        }
        else {
            $club_id = $this->vvEnv->getMIOIdByShortkey($input["shortkey"]);
            if( $club_id === FALSE ) {
                abort( 400, 'Invalid shortkey' );
            }
        }

        // Call VV function
        $aData = $this->vvEnv->vvGetClubInfo($club_id);

        if( $aData === FALSE ) {
            abort(500, 'Database error' );
        }
        elseif( empty($aData) ) {
            abort(404, 'No data found' );
        }

        $aData["updowngradeDate"] = "";
        $aData["cancelDate"] = "";
        $aData["changeStateDate"] = "";
        $aData["deleteDate"] = "";
        $aActions = (new Delayed)->getClubActions($this->vvEnv->getDatabaseId(), $this->vvEnv->getClubId());
        foreach( $aActions as $action ) {
            switch ($action->type) {
                case DelayedActions::ACTION_TYPE_UPDOWNGRADE:
                    $aData["updowngradeDate"] = $action->action_date->format('d.m.Y');
                    break;
                case DelayedActions::ACTION_TYPE_CANCEL:
                    $aData["cancelDate"] = $action->action_date->format('d.m.Y');
                    break;
                case DelayedActions::ACTION_TYPE_CHANGE:
                    $aData["changeStateDate"] = $action->action_date->format('d.m.Y');
                    break;
                case DelayedActions::ACTION_TYPE_DELETE:
                    $aData["deleteDate"] = $action->action_date->format('d.m.Y');;
                    break;
            }

        }


        return response()->json( $aData );
    }
}
