<?php

/**
 * verein.cloud Web VV API
 * ## Info ## API that allows data exchange with verein.cloud web application (VV)  ## Authentication ## Authentication for this API works through Keycloak where you will need to log in via OAuth and obtain a Java Web Token (JWT).  This token must then be passed to all calls in the Authorization Bearer header.  The API calls will return HTTP 401 when the token is not present or not valid.  The following image visualizes the process:  <img src=\"https://dev.comciencia.cl/api_auth.jpg\" width=\"800px\">
 *
 * The version of the OpenAPI document: 1.0.5
 * Contact: mheumann@comciencia.cl
 *
 * NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).
 * https://openapi-generator.tech
 * Do not edit the class manually.
 */


namespace App\Http\Controllers;

use Illuminate\Support\Facades\Request;
use App\Models\Databases;
use App\Services\VVInterface;

class ClubsApi extends Controller
{
    /**
     * Injected VVInterface for communication with the VV database
     * @var VVInterface
     */
    protected $vvEnv;

    /**
     * Constructor
     */
    public function __construct( VVInterface $vvEnv)
    {
        $this->vvEnv = $vvEnv;
    }

    /**
     * Operation getClubInfo
     *
     * Delivers information about the specified club.
     *
     * @return Http response
     */
    public function getClubInfo()
    {
        $input = Request::all();

        // Call VV function
        $aData = $this->vvEnv->vvGetClubData($this->vvEnv->getDatabaseId());

        if( $aData === FALSE ) {
            abort(500, 'Database error' );
        }
        elseif( empty($aData) ) {
            abort(404, 'No data found' );
        }

        return response()->json( $aData );
    }

    /**
     * Operation getListOfLetterSalutations
     *
     * Return a list of all available letter salutations for the specified club..
     *
     *
     * @return Http response
     */
    public function getListOfLetterSalutations()
    {
        // Call VV function
        $aData = $this->vvEnv->vvGetLetterSalutations();

        if( $aData === FALSE ) {
            abort(500, 'Database error' );
        }
        elseif( empty($aData) ) {
            abort(404, 'No data found' );
        }

        return response()->json( $aData );
    }

    /**
     * Operation getListOfMaritalStatus
     *
     * Return a list of all available marital status values for the specified club..
     *
     *
     * @return Http response
     */
    public function getListOfMaritalStatus()
    {
        $input = Request::all();

        // Call VV function
        $aData = $this->vvEnv->vvGetMaritalStatus();

        if( $aData === FALSE ) {
            abort(500, 'Database error' );
        }
        elseif( empty($aData) ) {
            abort(404, 'No data found' );
        }

        return response()->json( $aData );
    }

    /**
     * Operation getListOfSalutations
     *
     * Return a list of all available salutations for the specified club..
     *
     *
     * @return Http response
     */
    public function getListOfSalutations()
    {
        $input = Request::all();

        // Call VV function
        $aData = $this->vvEnv->vvGetSalutations();

        if( $aData === FALSE ) {
            abort(500, 'Database error' );
        }
        elseif( empty($aData) ) {
            abort(404, 'No data found' );
        }

        return response()->json( $aData );
    }

    /**
     * Operation getListOfTitles
     *
     * Return a list of all available titles for the specified club..
     *
     *
     * @return Http response
     */
    public function getListOfTitles()
    {
        // Call VV function
        $aData = $this->vvEnv->vvGetTitles();

        if( $aData === FALSE ) {
            abort(500, 'Database error' );
        }
        elseif( empty($aData) ) {
            abort(404, 'No data found' );
        }

        return response()->json( $aData );
    }

}
