<?php

/**
 * verein.cloud Web VV API
 *
 * Contact: mheumann@comciencia.cl
 *
 * NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).
 * https://openapi-generator.tech
 * Do not edit the class manually.
 */

namespace App\Http\Controllers;

use App\Models\Databases;
use App\Services\VVInterface;
use App\Rules\ValidArrayVal;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Request;
use Illuminate\Support\Facades\Validator;
use Symfony\Component\VarDumper\Cloner\Data;

class UsersApi extends Controller
{
    /**
     * Injected VVInterface for communication with the VV database
     * @var VVInterface
     */
    protected $vvEnv;

    /**
     * Constructor
     */
    public function __construct( VVInterface $vvEnv)
    {
        $this->vvEnv = $vvEnv;
    }

    public function handleUnknown()
    {
        return response()->json([
            'message' => 'Page Not Found. If error persists, contact provider'], 404);
    }

    /**
     * Operation getMemberInfo
     *
     * Delivers stored address info on passed member.
     *
     * @return Http response
     */
    public function getMemberInfo()
    {
        $input = Request::all();

        $member_id = $input['member_id'];

        // Call VV function
        $aData = $this->vvEnv->vvGetMitgliedsData($member_id);

        if ($aData === FALSE) {
            abort(500, 'Database error');
        }
        elseif (empty($aData)) {
            abort(404, 'No data found');
        }

        return response()->json($aData);
    }

    /**
     * Operation getMemberPhoto
     *
     * Delivers picture of the passed member, if available.
     *
     * @return Http response
     */
    public function getMemberPhoto()
    {
        $input = Request::all();

        $member_id = $input['member_id'];

        // Call VV function
        $mimeType = "";
        $strImage = $this->vvEnv->vvGetMitgliedsImage($member_id, $mimeType);

        if ($strImage === FALSE) {
            abort(500, 'Database error');
        }
        elseif (empty($strImage)) {
            $defaultImgPath = resource_path() . '/images/profile.png';
            $mimeType = "image/png";
            $strImage = file_get_contents($defaultImgPath);
        }

        $parts = explode( "/", $mimeType );
        $filename = "member_" . $this->vvEnv->getDatabaseId() . $this->vvEnv->getClubId() . $member_id . "." . $parts[1];

        return response($strImage)
            ->header('Content-Type', $mimeType)
            ->header('Pragma','public')
            ->header('Content-Disposition','inline; filename="'.$filename.'"')
            ->header('Cache-Control','max-age=60, must-revalidate');
    }

    /**
     * Operation getMemberPhotos
     *
     * Delivers the pictures of the specified members in a batch.
     *
     * @return \Illuminate\Http\Response|\Larave\Lumen\Http\ResponseFactory
     */
   public function getMemberPhotos()
    {
        $input = Request::all();

        $member_ids = $input['member_ids'];
        $aReturnData = array();

        // Call VV function
        foreach( $member_ids as $member_id) {
            $mimeType = "";
            $strImage = $this->vvEnv->vvGetMitgliedsImage($member_id, $mimeType);

            if (empty($strImage) || $strImage === FALSE) {
                // Show default image in case of any error during image retrieval
                $defaultImgPath = resource_path() . '/images/profile.png';
                $mimeType = "image/png";
                $strImage = file_get_contents($defaultImgPath);
            }

            $parts = explode("/", $mimeType);
            $filename = "member_" . $this->vvEnv->getDatabaseId() . $this->vvEnv->getClubId() . $member_id . "." . $parts[1];

            $oReturn = new \stdClass();
            $oReturn->member_id = $member_id;
            $oReturn->filename = $filename;
            $oReturn->photo = base64_encode($strImage);

            $aReturnData[] = $oReturn;
        }

        return response()->json($aReturnData);
    }

    /**
     * Operation getProfileInfo
     *
     * Delivers profile info about the passed member..
     *
     * @return Http response
     */
    public function getProfileInfo()
    {
        $input = Request::all();

        $member_id = $input['member_id'];

        // Make sure that the requested member is a part of this club.
        // This is done here and not in the middleware, because we are using the club middleware for this call,
        // because we don't want user permissions validation in this case (the call is allowed for any user of the club).
        if (!$this->vvEnv->vvIsMemberInClub($member_id)) {
            abort(404, 'No data found');
        }

        $aData = $this->vvEnv->vvGetProfileData($member_id);

        if ($aData === FALSE) {
            abort(500, 'Database error');
        }
        elseif (empty($aData)) {
            abort(404, 'No data found');
        }

        return response()->json($aData);
    }

    /**
     * Operation updateBankData
     *
     * Modifies a member's bank data.
     *
     *
     * @return Http response
     */
    public function updateBankData()
    {
        $input = Request::all();

        $member_id = $input['member_id'];

        if (!isset($input['iban'])) {
            throw new \InvalidArgumentException('Missing the required parameter $iban when calling updateBankData');
        }
        if (strlen($input['iban']) > 34) {
            throw new \InvalidArgumentException('invalid length for $iban when calling updateBankData, must be smaller than or equal to 34.');
        }
        $iban = $input['iban'];

        if (isset($input['bic'])) {
            if (strlen($input['bic']) > 11) {
                throw new \InvalidArgumentException('invalid length for $bic when calling updateBankData, must be smaller than or equal to 8.');
            }
            $bic = $input['bic'];
        }
        else {
            $bic = "";
        }

        if (isset($input['name'])) {
            if (strlen($input['name']) > 255) {
                throw new \InvalidArgumentException('invalid length for $name when calling updateBankData, must be smaller than or equal to 255.');
            }
            $name = $input['name'];
        }
        else {
            $name = "";
        }

        if (count($input) <= 3) {
            abort(400, 'No data sent');
        }

        // Validate input data
        $this->validate(Request::instance(), [
            'iban' => ['required', 'string', 'min:1', 'max:34'],
            'bic' => ['sometimes', 'string', 'min:1', 'max:11'],
            'name' => ['sometimes', 'string', 'min:1', 'max:255', 'regex:/^[\s,.\'\-\pL]+$/u'],
        ]);

        $countryCode = strtoupper(substr($iban, 0, 2));
        if ($countryCode == 'DE') {
            // Validate IBAN
            $aBankData = $this->vvEnv->vvCheckBankdata($iban, TRUE);

            if ($aBankData["status"] == "fail") {
                abort(404, 'IBAN invalid: ' . $aBankData["msg"]);
            }
            else {
                $aData = $aBankData["data"];
            }
        }
        else {
            $aData = array(
                "iban" => $iban,
                "bic" => $bic,
                "name" => $name
            );
        }

        try {
            $iCode = $this->vvEnv->vvUpdateBankInfo($member_id, $aData);
            switch ($iCode) {
                case 1:
                    abort(401, 'Specified member not found');
                    break;
                case 3:
                    abort(400, 'Error when saving member data');
                    break;
                default:
            }
        }
        catch (\Exception $exception) {
            abort(500, $exception->getMessage());
        }

        return response("OK", 200);
    }

    /**
     * Operation getDeptsContribsInfo
     *
     * Delivers information about the departments and contributions the member participates in.
     *
     *
     * @return Http response
     */
    public function getDeptsContribsInfo()
    {
        $input = Request::all();

        $member_id = $input['member_id'];

        // Call VV function
        $aData = $this->vvEnv->vvGetAbtBeitraege($member_id);

        if ($aData === FALSE) {
            abort(500, 'Database error');
        }
        elseif (empty($aData)) {
            abort(404, 'No data found');
        }

        return response()->json($aData);
    }

    /**
     * Operation updateMember
     *
     * Modifies member info.
     *
     *
     * @return Http response
     */
    public function updateMember()
    {
        $input = Request::all();

        $member_id = $input['member_id'];

        if (count($input) <= 3) {
            abort(400, 'No data sent');
        }

        $aTitles = $this->vvEnv->vvGetTitles(TRUE);
        $aSalutations = $this->vvEnv->vvGetSalutations(TRUE);
        $aLetterSalutations = $this->vvEnv->vvGetLetterSalutations(TRUE);
        $aMaritalStati = $this->vvEnv->vvGetMaritalStatus(TRUE);
        $aCountries = $this->vvEnv->vvGetCountryList(TRUE);
        $aCountries = array_column($aCountries, "value", "code");

        // Validate input data
        $this->validate(Request::instance(), [
            'lastName' => ['string', 'min:1', 'max:255', 'regex:/^[0-9\s,.\'\-\pL]+$/u'],
            'firstName' => ['string', 'min:1', 'max:50', 'regex:/^[\s,.\'\-\pL]+$/u'],
            'birthDate' => 'date_format:Y-m-d',
            'titleId' => ['integer', 'min:1', new ValidArrayVal($aTitles)],
            'salutationId' => ['integer', 'min:1', new ValidArrayVal($aSalutations)],
            'letterSalutationId' => ['integer', 'min:1', new ValidArrayVal($aLetterSalutations)],
            'maritalStatusId' => ['integer', 'min:1', new ValidArrayVal($aMaritalStati)],
            'gender' => 'string|max:1|in:M,m,F,f,B,b,D,d',
            'pobox' => ['sometimes', 'string', 'min:1', 'max:50', 'regex:/^[a-z0-9\- ]+$/iu'],
            'poboxPostCode' => ['sometimes', 'string', 'min:1', 'max:15', 'regex:/^[a-z0-9\- ]+$/iu'],
            'street' => ['string', 'min:1', 'max:100', 'regex:/^[0-9\s,.\'\-\pL]+$/u'],
            'street2' => ['string', 'min:1', 'max:100', 'regex:/^[0-9\s,.\'\-\pL]+$/u'],
            'poboxCity' => ['sometimes', 'string', 'min:1', 'max:50', 'regex:/^[0-9\s,.\'\-\pL]+$/u'],
            'postCode' => ['sometimes', 'string', 'min:1', 'max:15', 'regex:/^[a-z0-9\- ]+$/iu'],
            'city' => ['string', 'min:1', 'max:50', 'regex:/^[0-9\s,.\'\-\pL]+$/u'],
            'countryCode' => ['string', 'min:2', 'max:2', new ValidArrayVal($aCountries)],
            'email' => 'email|max:255',
            'phone' => ['string', 'min:1', 'max:100', 'regex:/^[0-9\-+\/\(\) ]+$/iu'],
            'emergencyContact' => ['string', 'min:1', 'max:100', 'regex:/^[0-9\s,.\'\-\pL]+$/u'],
            'emergencyPhone' => ['string', 'min:1', 'max:100', 'regex:/^[0-9\-+\/\(\) ]+$/iu'],
            'shareBirthday' => ['boolean'],
        ]);

        try {
            $iCode = $this->vvEnv->vvUpdateMemberInfo($member_id, $input);
            switch ($iCode) {
                case 1:
                    abort(401, 'Specified member not found');
                    break;
                case 2:
                    abort(400, 'Requested change would create member duplicate');
                    break;
                case 3:
                    abort(400, 'Error when saving member data');
                    break;
                default:
            }
        }
        catch (\Exception $exception) {
            abort(500, $exception->getMessage());
        }

        return response("OK", 200);
    }

    /**
     * Operation uploadMemberPhoto
     *
     * Allows upload of a member's photo..
     *
     * @return Http response
     */
    public function uploadMemberPhoto()
    {
        $input = Request::all();

        $member_id = $input['member_id'];

        if( empty($input['image_file']) ) {
            abort(406, 'Image is empty');
        }

        $validator = Validator::make($input, [
            'image_file' => 'required|is_image|max:10485760'
        ]);
        if ($validator->fails()) {
            abort(406, 'Image is too large or invalid type (only JPG, GIF, PNG and SVG allowed)');
        }

        $image = base64_decode($input['image_file']);
        $f = finfo_open();
        $mimeType = finfo_buffer($f, $image, FILEINFO_MIME_TYPE);

        $iResult = $this->vvEnv->vvSaveMemberPhoto($member_id, $mimeType, $input['image_file']);
        if (!$iResult) {
            abort(400, 'Error when saving member photo');
        }

        return response("OK", 200);
    }

}
