<?php

namespace App\Http\Middleware;

use Closure;
use Illuminate\Http\Response;
use Illuminate\Support\Facades\Log;
use GuzzleHttp\Utils;

class APILoggingMiddleware
{
    /**
     * Handle an incoming request.
     *
     * @param  \Illuminate\Http\Request  $request
     * @param  \Closure  $next
     * @return mixed
     */
    public function handle($request, Closure $next)
    {
        return $next($request);
    }

    /**
     * Completed request.
     *
     * @param  \Illuminate\Http\Request  $request
     * @param  Response  $response
     * @return mixed
     */
    public function terminate ($request, $response)
    {
        $url = $request->path();
        $status = $response->status();
        $method = $request->method();
        $ip = $request->ip();
        $log = [
            'request' => [
                'raw' => $request->getContent(),
                'header' => $request->header(),
                'body' => $request->all()
            ],
        ];

        // Replace bearer for security and because it is too long
        if( isset($log["request"]["header"]["authorization"]) &&
            substr($log["request"]["header"]["authorization"][0], 0, 6) == "Bearer" ) {
            $log["request"]["header"]["authorization"][0] = "Bearer [...removed...]";
        }

        if (env('API_LOG_RESPONSE')) {
            $log['response'] = [
                'body' => $response->content(),
            ];
        }

        try {
            $strLogMsg = "[$status] [$ip] $method $url " . Utils::jsonEncode($log);
        }
        catch (\Exception $e) {
            // Binary response can't be logged
            $strLogMsg = "[$status] [$ip] $method $url REQUEST " . Utils::jsonEncode($log['request']);
            $strLogMsg .= " - RESPONSE BINARY";
        }

        if ($status >= 500) {
             Log::error($strLogMsg);
        }
        elseif ($status >= 400) {
             Log::warning($strLogMsg);
        }
        else {
             Log::info($strLogMsg);
        }
    }
}
