<?php

namespace App\Http\Middleware;

use App\Services\VVInterfaceCRM;
use Closure;
use Illuminate\Support\Facades\App;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\Request;
use Illuminate\Support\Facades\Validator;

class ValidateCRMClub
{
    /**
     * Handle an incoming request.
     *
     * @param  \Illuminate\Http\Request  $request
     * @param  \Closure  $next
     * @return mixed
     */
    public function handle($request, Closure $next)
    {
        $input = Request::all();

        $validator = Validator::make($input, [
            'database_id' => ['required', 'numeric'],
            'club_id' => ['required', 'numeric'],
        ]);
        if ($validator->fails()) {
            abort(422, $validator->errors()->first() );
        }

        if (!isset($input['database_id'])) {
            abort( 400, 'Missing the required parameter $database_id' );
        }
        if ($input['database_id'] < 1) {
            abort( 400, 'invalid value for $database_id, must be bigger than or equal to 1.' );
        }

        if (!isset($input['club_id'])) {
            abort(400, 'Missing the required parameter $club_id');
        }
        if ($input['club_id'] < 1) {
            abort(400, 'invalid value for $club_id, must be bigger than or equal to 1.');
        }

        $database_id = (int)$input['database_id'];
        $club_id = (int)$input['club_id'];

        Log::debug( "Database ID $database_id, Club ID $club_id" );

        // Register the singleton with dynamic params
        App::singleton(VVInterfaceCRM::class, function () use ($database_id, $club_id) {
            return new VVInterfaceCRM($database_id, $club_id);
        });

        return $next($request);
    }
}
