<?php

namespace App\Http\Middleware;

use App\Services\VVInterface;
use Closure;
use Illuminate\Support\Facades\App;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\Request;
use Illuminate\Support\Facades\Validator;

class ValidateClub
{
    /**
     * Handle an incoming request.
     *
     * @param  \Illuminate\Http\Request  $request
     * @param  \Closure  $next
     * @return mixed
     */
    public function handle($request, Closure $next)
    {
        $input = Request::all();

        $validator = Validator::make($input, [
            'database_id' => ['required', 'numeric'],
            'club_id' => ['required', 'numeric'],
        ]);
        if ($validator->fails()) {
            abort(422, $validator->errors()->first() );
        }

        if (!isset($input['database_id'])) {
            abort( 400, 'Missing the required parameter $database_id when calling getMemberInfo' );
        }
        if ($input['database_id'] < 1) {
            abort( 400, 'invalid value for $database_id when calling getMemberInfo, must be bigger than or equal to 1.' );
        }

        if (!isset($input['club_id'])) {
            abort(400, 'Missing the required parameter $club_id when calling getMemberInfo');
        }
        if ($input['club_id'] < 1) {
            abort(400, 'invalid value for $club_id when calling getMemberInfo, must be bigger than or equal to 1.');
        }
        $club_id = $input['club_id'];

        $token = json_decode(Auth::token());

        // Club ID must be same as the one the member belongs to
        if( $club_id != $token->attributes->vv_club_id ) {
            abort( 403, "Attempted access to forbidden resource (club)" );
        }

        $database_id = (int)$input['database_id'];
        $club_id = (int)$input['club_id'];

        // Register the singleton with dynamic params
        App::singleton(VVInterface::class, function () use ($database_id, $club_id) {
            return new VVInterface($database_id, $club_id);
        });

        return $next($request);
    }
}
