<?php

namespace App\Http\Middleware;

use Closure;
use Illuminate\Support\Facades\App;
use Illuminate\Support\Facades\Auth;
use App\Services\VVInterface;
use Illuminate\Support\Facades\Request;
use App\Models\Databases;
use Illuminate\Support\Facades\Validator;

class ValidateUser
{
    /**
     * Handle an incoming request.
     *
     * @param  \Illuminate\Http\Request  $request
     * @param  \Closure  $next
     * @return mixed
     */
    public function handle($request, Closure $next)
    {
       $input = Request::all();

       $validator = Validator::make($input, [
            'database_id' => ['required', 'numeric'],
            'club_id' => ['required', 'numeric'],
            'member_id' => ['required', 'numeric'],
        ]);
        if ($validator->fails()) {
            abort(422, $validator->errors()->first() );
        }

        if (!isset($input['database_id'])) {
            abort( 400, 'Missing the required parameter $database_id when calling getMemberInfo' );
        }
        if ($input['database_id'] < 1) {
            abort( 400, 'invalid value for $database_id when calling getMemberInfo, must be bigger than or equal to 1.' );
        }
        $database_id = (int)$input['database_id'];

        if (!isset($input['club_id'])) {
            abort(400, 'Missing the required parameter $club_id when calling getMemberInfo');
        }
        if ($input['club_id'] < 1) {
            abort(400, 'invalid value for $club_id when calling getMemberInfo, must be bigger than or equal to 1.');
        }
        $club_id = (int)$input['club_id'];

        if (!isset($input['member_id'])) {
            abort(400, 'Missing the required parameter $member_id when calling getMemberInfo');
        }
        if ($input['member_id'] < 1) {
            abort(400, 'invalid value for $member_id when calling getMemberInfo, must be bigger than or equal to 1.');
        }
        $member_id = (int)$input['member_id'];

        $token = json_decode(Auth::token());

        // Club ID must be same as the one the member belongs to
        if( $club_id != $token->attributes->vv_club_id ) {
            abort( 403, "Attempted access to forbidden resource (club)" );
        }

        // For non-admins, member ID must be same as the current user's id
        $aKeycloakConfig = config( "keycloak" );
        $aUseConfig = $aKeycloakConfig;

        if( !empty($aKeycloakConfig['clients']) && file_exists(base_path($aKeycloakConfig['clients'])) ) {
            $jsonFile = file_get_contents(base_path($aKeycloakConfig['clients']));
            // Overwrite clients section with JSON
            $json = json_decode($jsonFile, true);
            if( !empty($json) ) {
                $aKeycloakConfig['clients'] = $json;
                $aUseConfig = $aKeycloakConfig['clients'][$token->azp];
            }
        }

        if( !Auth::hasRole( $aUseConfig["allowed_resources"], 'admin' ) &&
            $member_id != $token->attributes->vv_member_id ) {
            abort( 403, "Attempted access to forbidden resource (member)" );
        }

        // Register the singleton with dynamic params
        App::singleton(VVInterface::class, function () use ($database_id, $club_id) {
            return new VVInterface($database_id, $club_id);
        });

        // Call VV function
        if( !App(VVInterface::class)->vvIsMemberInClub( $member_id ) ) {
            abort(404, 'No data found' );
        }

        return $next($request);
    }
}
