<?php

namespace App\Jobs;

use App\Services\VVInterfaceCRM;
use App\Models\Databases;
use Illuminate\Support\Facades\Log;
use Throwable;

class ChangeClubs extends Job
{
    protected $database_id;
    protected $club_id;
    protected $new_state;

    /**
     * Create a new job instance.
     * @param  int  $database_id
     * @param  int  $club_id
     * @param  int  $new_state
     * @return void
     */
    public function __construct( $database_id, $club_id, $new_state )
    {
        $this->database_id = $database_id;
        $this->club_id = $club_id;
        $this->new_state = $new_state;
    }

    /**
     * Execute the job.
     *
     * @return void
     */
    public function handle()
    {
        Log::debug( "Job ChangeClubs, time: " . date("d.m.Y H:i:s") .
            ", processing: $this->database_id, $this->club_id" );

        $dbdata = Databases::where('databaseid', $this->database_id);

        if( !empty($dbdata) && !is_null($dbdata->first()) ) {
            $dbdata = $dbdata->first()->toArray();
        }

        $vvEnv = new VVInterfaceCRM($dbdata);
        $vvEnv->disableEnableClub( $this->database_id, $this->club_id, ($this->new_state == 1) );
        Log::debug( "ChangeClubs, job for {$this->club_id} (DB {$this->database_id}) successful." );
    }

    /**
     * Handle a job failure.
     */
    public function failed(Throwable $exception): void
    {
        Log::debug( "Job failed because " . $exception->getMessage() );
    }
}
