<?php

namespace App\Jobs;

use App\Services\VVInterfaceCRM;
use App\Models\Databases;
use Illuminate\Support\Facades\Log;

class DeleteClubs extends Job
{
    protected $database_id;
    protected $club_id;

    /**
     * Create a new job instance.
     * @param  int  $database_id
     * @param  int  $club_id
     * @return void
     */
    public function __construct( $database_id, $club_id )
    {
        $this->database_id = $database_id;
        $this->club_id = $club_id;
    }

    /**
     * Execute the job.
     *
     * @return void
     */
    public function handle()
    {
        Log::debug( "Job DeleteClubs, time: " . date("d.m.Y H:i:s") .
            ", processing: $this->database_id, $this->club_id" );

        $dbdata = Databases::where('databaseid', $this->database_id);

        if( !empty($dbdata) && !is_null($dbdata->first()) ) {
            $dbdata = $dbdata->first()->toArray();
        }

        $vvEnv = new VVInterfaceCRM($dbdata);
        $bSuccess = $vvEnv->deleteClub( $this->database_id, $this->club_id );

        if( !$bSuccess ) {
            Log::debug( "DeleteClubs, job for {$this->club_id} (DB {$this->database_id}) failed." );
        }
        else {
            Log::debug( "DeleteClubs, job for {$this->club_id} (DB {$this->database_id}) executed successfully, Club deleted." );
        }
    }
}
