<?php

namespace App\Jobs;

use App\Services\VVInterfaceCRM;
use App\Models\Databases;
use Illuminate\Support\Facades\Log;

class UpDowngradeClubs extends Job
{
    protected $database_id;
    protected $club_id;
    protected $new_state;

    /**
     * Create a new job instance.
     * @param  int  $database_id
     * @param  int  $club_id
     * @param  int  $new_state 0 = downgrade, 1 = upgrade
     * @return void
     */
    public function __construct( $database_id, $club_id, $new_state )
    {
        $this->database_id = $database_id;
        $this->club_id = $club_id;
        $this->new_state = $new_state;
    }

    /**
     * Execute the job.
     *
     * @return void
     */
    public function handle()
    {
        Log::debug( "Job UpDowngradeClubs, time: " . date("d.m.Y H:i:s") .
            ", processing: $this->database_id, $this->club_id, $this->new_state" );

        $dbdata = Databases::where('databaseid', $this->database_id);

        if( !empty($dbdata) && !is_null($dbdata->first()) ) {
            $dbdata = $dbdata->first()->toArray();
        }

        $vvEnv = new VVInterfaceCRM($dbdata);
        if( $vvEnv->updowngradeClub( $this->club_id, (boolean)$this->new_state ) ) {
            Log::debug( "UpDowngradeClubs, job for {$this->club_id} (DB {$this->database_id}) " .
                ($this->new_state ? "up" : "down") . " succeeded." );
        }
        else {
            Log::debug( "UpDowngradeClubs, job for {$this->club_id} (DB {$this->database_id}) " .
                ($this->new_state ? "up" : "down") . " failed." );
        }
    }
}
