<?php

namespace App\Models;

use Illuminate\Database\Eloquent\Model;

class Login extends Model
{
    protected $table = 'LOGIN';
    protected $primaryKey = 'LOGINID';
    public $timestamps = false;

    public function insertRecord($p_iDatabaseId, $p_strShortkey, $p_iActive)
    {
        $this->databaseid = $p_iDatabaseId;
        $this->shortkey = $p_strShortkey;
        $this->active = $p_iActive;
        $this->activemessage = "";
        $this->save();
    }

    public function isShortkeyUnique($p_strShortkey, $p_iExcludeId = -1)
    {
        if ( $p_iExcludeId > -1 ) {
            $count = $this->where([
                ['shortkey', '=', $p_strShortkey],
                ['loginid', '<>', $p_iExcludeId]
            ])->count();
        }
        else {
            $count = $this->where('shortkey', $p_strShortkey)->count();
        }

        if( $count > 0 ) {
            return FALSE;
        }
        else {
            return TRUE;
        }
    }

    /**
     * @param $p_strOrgName
     * @return string
     */
    public function suggestShortkey( $p_strOrgName )
    {
        $aForbiddenNames = array(
            'kz', 'hj', 'ns', 'sa', 'sd', 'ss', 'npd', 'nsu'
        );

        // Nimm jedes Wort: zwei Zeichen übernehmen, längere Wörter nur Anfangsbuchstaben
        $tmpname = str_replace( array('.', '-'), array('', ' '), $p_strOrgName );  // remove dots and replace dashes with blanks
        $words = explode(' ', transliterator_transliterate('Any-Latin; Latin-ASCII', $tmpname));  // parse text into words (separators=blanks)

        // If organisation name starts with a word of up to 4 uppercase letters, assume it's an acronym and use it
        $name = '';
        foreach ($words as $word) {
            if (ctype_upper($word) && strlen($word) < 4) {
              $name .= $word;
            }
            else {
              $name .= substr($word, 0, 1);
            }
        }

        $i = 1;
        $name = strtolower($name);
        $name = preg_replace('/[^a-z0-9]/','', $name); // Remove non alphanumeric characters
        $temp = $name;
        while (TRUE) {
            if (!in_array($temp, $aForbiddenNames) && $this->isShortkeyUnique( $temp )) {
               return $temp;
            }
            $i++;
            $temp = $name . $i;
        }
    }

    /**
     * Get the database id for the passed shortkey.
     *
     * @param string $p_strShortkey Shortkey used to find club
     * @return integer databaseid of club identified by passed shortkey.
     */
    public function getDatabaseId($p_strShortkey)
    {
       return $this->where('shortkey', $p_strShortkey)->value('databaseid');
    }
}
