<?php

namespace App\Services;

use App\Exceptions\KeycloakException;
use App\Models\Databases;
use DateTime;
use Illuminate\Support\Facades\Log;

class Keycloak
{
    // Parameters
    private static $KEYCLOAKAPI_REALM                         = NULL;
    private static $KEYCLOAKAPI_URL                           = NULL;
    private static $KEYCLOAKAPI_ADMIN_USERNAME                = NULL;
    private static $KEYCLOAKAPI_ADMIN_PASSWORD                = NULL;
    private static $KEYCLOAKAPI_ADMIN_CLIENT_ID               = NULL;
    private static $KEYCLOAKAPI_USER_CLIENT_ID                = NULL;
    private static $KEYCLOAKAPI_USER_CLIENT_SECRET            = NULL;

    // Roles
    private static $KEYCLOAKAPI_REALM_ROLE_CLUB_MEMBERS       = NULL;
    private static $KEYCLOAKAPI_REALM_ROLE_CLUB_ADMINS        = NULL;
    private static $KEYCLOAKAPI_REALM_ROLE_CLUB_GUESTS        = NULL;
    private static $KEYCLOAKAPI_CLIENT_ROLE_ADMIN             = NULL;
    private static $KEYCLOAKAPI_CLIENT_ROLE_USER              = NULL;
    private static $KEYCLOAKAPI_CLIENT_ROLE_GUEST             = NULL;
    private static $KEYCLOAKAPI_CLIENT_ROLE_FUNCTIONARY       = NULL;
    private static $KEYCLOAKAPI_CLIENT_ROLE_SECRETARY         = NULL;
    private static $KEYCLOAKAPI_CLIENT_ROLE_EDITOR            = NULL;
    private static $KEYCLOAKAPI_CLIENT_ROLE_MEMBER_LIGHT      = NULL;
    private static $KEYCLOAKAPI_CLIENT_ROLE_ADMIN_LIGHT       = NULL;

    // Roles Descriptions
    private static $KEYCLOAKAPI_DESC_REALM_ROLE_CLUB_MEMBERS  = NULL;
    private static $KEYCLOAKAPI_DESC_REALM_ROLE_CLUB_ADMINS   = NULL;
    private static $KEYCLOAKAPI_DESC_REALM_ROLE_CLUB_GUESTS   = NULL;
    private static $KEYCLOAKAPI_DESC_CLIENT_ROLE_ADMIN        = NULL;
    private static $KEYCLOAKAPI_DESC_CLIENT_ROLE_USER         = NULL;
    private static $KEYCLOAKAPI_DESC_CLIENT_ROLE_GUEST        = NULL;
    private static $KEYCLOAKAPI_DESC_CLIENT_ROLE_FUNCTIONARY  = NULL;
    private static $KEYCLOAKAPI_DESC_CLIENT_ROLE_SECRETARY    = NULL;
    private static $KEYCLOAKAPI_DESC_CLIENT_ROLE_EDITOR       = NULL;
    private static $KEYCLOAKAPI_DESC_CLIENT_ROLE_MEMBER_LIGHT = NULL;
    private static $KEYCLOAKAPI_DESC_CLIENT_ROLE_ADMIN_LIGHT  = NULL;

    // Attributes
    private static $KEYCLOAKAPI_ATTRIBUTE_DBID                = NULL;
    private static $KEYCLOAKAPI_ATTRIBUTE_MIOID               = NULL;
    private static $KEYCLOAKAPI_ATTRIBUTE_ADRID               = NULL;
    private static $KEYCLOAKAPI_ATTRIBUTE_PHONE               = NULL;
    private static $KEYCLOAKAPI_ATTRIBUTE_UPDDA               = NULL;

    //Tokens
    private static $KEYCLOAKAPI_ACCESS_TOKEN_VALUE            = NULL;
    private static $KEYCLOAKAPI_ACCESS_TOKEN_EXPIRES_AT       = NULL;
    private static $KEYCLOAKAPI_REFRESH_TOKEN_VALUE           = NULL;
    private static $KEYCLOAKAPI_REFRESH_TOKEN_EXPIRES_AT      = NULL;

    // Common Time : for register/updates operations
    private static $KEYCLOAKAPI_TIME                          = NULL;

    //Time outs
    private static $KEYCLOAKAPI_CONNECTTIMEOUT                = 5;
    private static $KEYCLOAKAPI_TIMEOUT                       = 30;

    private static $error_codes=array(
                    1  => 'CURLE_UNSUPPORTED_PROTOCOL',
                    2  => 'CURLE_FAILED_INIT',
                    3  => 'CURLE_URL_MALFORMAT',
                    4  => 'CURLE_URL_MALFORMAT_USER',
                    5  => 'CURLE_COULDNT_RESOLVE_PROXY',
                    6  => 'CURLE_COULDNT_RESOLVE_HOST',
                    7  => 'CURLE_COULDNT_CONNECT',
                    8  => 'CURLE_FTP_WEIRD_SERVER_REPLY',
                    9  => 'CURLE_REMOTE_ACCESS_DENIED',
                    11 => 'CURLE_FTP_WEIRD_PASS_REPLY',
                    13 => 'CURLE_FTP_WEIRD_PASV_REPLY',
                    14 => 'CURLE_FTP_WEIRD_227_FORMAT',
                    15 => 'CURLE_FTP_CANT_GET_HOST',
                    17 => 'CURLE_FTP_COULDNT_SET_TYPE',
                    18 => 'CURLE_PARTIAL_FILE',
                    19 => 'CURLE_FTP_COULDNT_RETR_FILE',
                    21 => 'CURLE_QUOTE_ERROR',
                    22 => 'CURLE_HTTP_RETURNED_ERROR',
                    23 => 'CURLE_WRITE_ERROR',
                    25 => 'CURLE_UPLOAD_FAILED',
                    26 => 'CURLE_READ_ERROR',
                    27 => 'CURLE_OUT_OF_MEMORY',
                    28 => 'CURLE_OPERATION_TIMEDOUT',
                    30 => 'CURLE_FTP_PORT_FAILED',
                    31 => 'CURLE_FTP_COULDNT_USE_REST',
                    33 => 'CURLE_RANGE_ERROR',
                    34 => 'CURLE_HTTP_POST_ERROR',
                    35 => 'CURLE_SSL_CONNECT_ERROR',
                    36 => 'CURLE_BAD_DOWNLOAD_RESUME',
                    37 => 'CURLE_FILE_COULDNT_READ_FILE',
                    38 => 'CURLE_LDAP_CANNOT_BIND',
                    39 => 'CURLE_LDAP_SEARCH_FAILED',
                    41 => 'CURLE_FUNCTION_NOT_FOUND',
                    42 => 'CURLE_ABORTED_BY_CALLBACK',
                    43 => 'CURLE_BAD_FUNCTION_ARGUMENT',
                    45 => 'CURLE_INTERFACE_FAILED',
                    47 => 'CURLE_TOO_MANY_REDIRECTS',
                    48 => 'CURLE_UNKNOWN_TELNET_OPTION',
                    49 => 'CURLE_TELNET_OPTION_SYNTAX',
                    51 => 'CURLE_PEER_FAILED_VERIFICATION',
                    52 => 'CURLE_GOT_NOTHING',
                    53 => 'CURLE_SSL_ENGINE_NOTFOUND',
                    54 => 'CURLE_SSL_ENGINE_SETFAILED',
                    55 => 'CURLE_SEND_ERROR',
                    56 => 'CURLE_RECV_ERROR',
                    58 => 'CURLE_SSL_CERTPROBLEM',
                    59 => 'CURLE_SSL_CIPHER',
                    60 => 'CURLE_SSL_CACERT',
                    61 => 'CURLE_BAD_CONTENT_ENCODING',
                    62 => 'CURLE_LDAP_INVALID_URL',
                    63 => 'CURLE_FILESIZE_EXCEEDED',
                    64 => 'CURLE_USE_SSL_FAILED',
                    65 => 'CURLE_SEND_FAIL_REWIND',
                    66 => 'CURLE_SSL_ENGINE_INITFAILED',
                    67 => 'CURLE_LOGIN_DENIED',
                    68 => 'CURLE_TFTP_NOTFOUND',
                    69 => 'CURLE_TFTP_PERM',
                    70 => 'CURLE_REMOTE_DISK_FULL',
                    71 => 'CURLE_TFTP_ILLEGAL',
                    72 => 'CURLE_TFTP_UNKNOWNID',
                    73 => 'CURLE_REMOTE_FILE_EXISTS',
                    74 => 'CURLE_TFTP_NOSUCHUSER',
                    75 => 'CURLE_CONV_FAILED',
                    76 => 'CURLE_CONV_REQD',
                    77 => 'CURLE_SSL_CACERT_BADFILE',
                    78 => 'CURLE_REMOTE_FILE_NOT_FOUND',
                    79 => 'CURLE_SSH',
                    80 => 'CURLE_SSL_SHUTDOWN_FAILED',
                    81 => 'CURLE_AGAIN',
                    82 => 'CURLE_SSL_CRL_BADFILE',
                    83 => 'CURLE_SSL_ISSUER_ERROR',
                    84 => 'CURLE_FTP_PRET_FAILED',
                    84 => 'CURLE_FTP_PRET_FAILED',
                    85 => 'CURLE_RTSP_CSEQ_ERROR',
                    86 => 'CURLE_RTSP_SESSION_ERROR',
                    87 => 'CURLE_FTP_BAD_FILE_LIST',
                    88 => 'CURLE_CHUNK_FAILED');

    public const KEYCLOAKAPI_BATCH_SIZE = 50;
    public const KEYCLOAKAPI_BATCH_PAUSE = 10;

    /**
     * Private utility function to query a rest service.
     * @param $p_strAdresse     string a URL to connect to.
     * @param $p_strMethode     string a METHOD used to connect (GET, POST, PUT, DELETE ...).
     * @param $p_strDaten       string optional, a set of variables/data to add to the body of the request.
     * @param $p_strContentType string optional, a content type that fits the body.
     * @param $p_strAccessToken string optional, an access token.
     * @return array
     */
    private static function rest_call($p_strAdresse, $p_strMethode, $p_strDaten = NULL, $p_strContentType = NULL, $p_strAccessToken = NULL) {
        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $p_strAdresse);
        curl_setopt($curl, CURLOPT_CONNECTTIMEOUT, self::$KEYCLOAKAPI_CONNECTTIMEOUT);
        curl_setopt($curl, CURLOPT_TIMEOUT, self::$KEYCLOAKAPI_TIMEOUT);
        curl_setopt($curl, CURLOPT_CUSTOMREQUEST, $p_strMethode);
        if ( isset($p_strContentType) ) {
            $head = ['Content-Type: ' . $p_strContentType];
        }
        if ( isset($p_strAccessToken) ) {
            if( !isset($head) ) {
                $head = array();
            }
            array_push($head, "Authorization: bearer " . $p_strAccessToken);
        }
        if ( isset($head) ) {
            curl_setopt($curl, CURLOPT_HTTPHEADER, $head);
        }
        if ( isset($p_strDaten) ) {
            curl_setopt($curl, CURLOPT_POSTFIELDS, $p_strDaten);
        }
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
        $headers = [];
        curl_setopt($curl, CURLOPT_HEADERFUNCTION,
            function($curl, $header) use (&$headers) {
                $len = strlen($header);
                $header = explode(':', $header, 2);
                if (count($header) < 2) // ignore invalid headers
                    return $len;
                $headers[strtolower(trim($header[0]))][] = trim($header[1]);
                return $len;
            }
        );
        $response = curl_exec($curl);
        $errnumbe = curl_errno($curl);
        if($errnumbe){
            //TIME OUT
            switch($errnumbe) {
                case 7:
                case 28:
                    return array( 'status'   => 408,
                                  'headers'  => array( 'Content-Length' => 32, 'Content-Type' => config('constants.keycloakSync.CONTENTTYPE_JSON') ),
                                  'response' => '{ "error": "HTTP 408 Time-Out" }' );
                default:
                    $errornum = 900 + $errnumbe;
                    $errortxt = str_replace('_',' ', str_replace('CURLE_', '', self::$error_codes[$errnumbe]));
                    $errortxt = sprintf('{ "error": "CURL ERROR %1$s - %2$s" }', $errnumbe, $errortxt);
                    return array( 'status'   => $errornum,
                                  'headers'  => array( 'Content-Length' => strlen($errortxt), 'Content-Type' => config('constants.keycloakSync.CONTENTTYPE_JSON') ),
                                  'response' => $errortxt );
            }
        }
        $status   = curl_getinfo($curl, CURLINFO_HTTP_CODE);
        return array( 'status'   => $status,
                      'headers'  => $headers,
                      'response' => $response );
    }

    /**
     * Private utility function to check whether keycloak parameters are set
     * @return bool
     */
    private static function keycloackCheckParameters() {
        if( self::$KEYCLOAKAPI_REALM                         === NULL ||
            self::$KEYCLOAKAPI_URL                           === NULL ||
            self::$KEYCLOAKAPI_ADMIN_USERNAME                === NULL ||
            self::$KEYCLOAKAPI_ADMIN_PASSWORD                === NULL ||
            self::$KEYCLOAKAPI_ADMIN_CLIENT_ID               === NULL ||
            self::$KEYCLOAKAPI_USER_CLIENT_ID                === NULL ||
            self::$KEYCLOAKAPI_USER_CLIENT_SECRET            === NULL ||
            self::$KEYCLOAKAPI_REALM_ROLE_CLUB_MEMBERS       === NULL ||
            self::$KEYCLOAKAPI_REALM_ROLE_CLUB_ADMINS        === NULL ||
            self::$KEYCLOAKAPI_REALM_ROLE_CLUB_GUESTS        === NULL ||
            self::$KEYCLOAKAPI_CLIENT_ROLE_ADMIN             === NULL ||
            self::$KEYCLOAKAPI_CLIENT_ROLE_USER              === NULL ||
            self::$KEYCLOAKAPI_CLIENT_ROLE_GUEST             === NULL ||
            self::$KEYCLOAKAPI_CLIENT_ROLE_FUNCTIONARY       === NULL ||
            self::$KEYCLOAKAPI_CLIENT_ROLE_SECRETARY         === NULL ||
            self::$KEYCLOAKAPI_CLIENT_ROLE_EDITOR            === NULL ||
            self::$KEYCLOAKAPI_CLIENT_ROLE_MEMBER_LIGHT      === NULL ||
            self::$KEYCLOAKAPI_CLIENT_ROLE_ADMIN_LIGHT       === NULL ||
            self::$KEYCLOAKAPI_DESC_REALM_ROLE_CLUB_MEMBERS  === NULL ||
            self::$KEYCLOAKAPI_DESC_REALM_ROLE_CLUB_ADMINS   === NULL ||
            self::$KEYCLOAKAPI_DESC_REALM_ROLE_CLUB_GUESTS   === NULL ||
            self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_ADMIN        === NULL ||
            self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_USER         === NULL ||
            self::$KEYCLOAKAPI_ATTRIBUTE_DBID                === NULL ||
            self::$KEYCLOAKAPI_ATTRIBUTE_MIOID               === NULL ||
            self::$KEYCLOAKAPI_ATTRIBUTE_ADRID               === NULL ||
            self::$KEYCLOAKAPI_ATTRIBUTE_PHONE               === NULL ||
            self::$KEYCLOAKAPI_ATTRIBUTE_UPDDA               === NULL) {
            return FALSE;
        }
        return TRUE;
    }

    /**
     * Private utility function to obtain an admin access token.
     * @return mixed in case of an error returns an array,
     *               in case of success it returns a string
     */
    private static function keycloackGetAdminToken() {
        if( !self::keycloackCheckParameters() ) {
            return array( 'ok'       => FALSE,
                          'message'  => 'KEYCLOAK PARAMETERS NOT SET',
                          'response' => '');
        }
        if( self::$KEYCLOAKAPI_ACCESS_TOKEN_VALUE      !== NULL &&
            self::$KEYCLOAKAPI_ACCESS_TOKEN_EXPIRES_AT !== NULL &&
            self::$KEYCLOAKAPI_ACCESS_TOKEN_EXPIRES_AT > new DateTime('now')
          ) {
            //If we are being called repeatedly in the same
            //program, we check if the previous obtained
            //admin token has expired, if not, reuse it.
            return self::$KEYCLOAKAPI_ACCESS_TOKEN_VALUE;
        }

        $keycloak_url = self::$KEYCLOAKAPI_URL;
        $adresse = $keycloak_url . "/realms/".self::$KEYCLOAKAPI_REALM."/protocol/openid-connect/token";
        $methode = "POST";
        if( self::$KEYCLOAKAPI_REFRESH_TOKEN_VALUE      !== NULL &&
            self::$KEYCLOAKAPI_REFRESH_TOKEN_EXPIRES_AT !== NULL &&
            self::$KEYCLOAKAPI_REFRESH_TOKEN_EXPIRES_AT > new DateTime('now')
          ) {
            //If we are being called repeatedly in the same
            //program, we check if the previous obtained
            //refresh token has expired, if not, we use it
            //to get a new admin/refresh token.
            $daten = http_build_query(array(
                'grant_type'    => 'refresh_token',
                'refresh_token' => self::$KEYCLOAKAPI_REFRESH_TOKEN_VALUE,
                'client_id'     => self::$KEYCLOAKAPI_ADMIN_CLIENT_ID
            ));
        }
        else{
            $daten = http_build_query(array(
                'grant_type' => 'password',
                'username'   => self::$KEYCLOAKAPI_ADMIN_USERNAME,
                'password'   => self::$KEYCLOAKAPI_ADMIN_PASSWORD,
                'client_id'  => self::$KEYCLOAKAPI_ADMIN_CLIENT_ID
            ));
        }
        $content_type = config('constants.keycloakSync.CONTENTTYPE_WWW_FORM_URLENCODED');

        //error_log( "DATEN " . print_r($daten, true) );

        $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, NULL);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'REST CALL ERROR',
                          'response' => $rest_response );
        }

        $rest_response = json_decode($rest_response['response']);

        if( !property_exists($rest_response, 'access_token') )
        {
            $error = 'NO ACCESS TOKEN IN RESPONSE';
            if( is_object( $rest_response ) && property_exists($rest_response, 'error') && property_exists($rest_response, 'error_description') )
            {
                $error = $rest_response->{'error_description'};
            }
            return array( 'ok'       => FALSE,
                          'message'  => $error,
                          'response' => $rest_response);
        }

        $expires                                   = (0 + $rest_response->{'expires_in'}) -3;
        self::$KEYCLOAKAPI_ACCESS_TOKEN_VALUE      = $rest_response->{'access_token'};
        self::$KEYCLOAKAPI_ACCESS_TOKEN_EXPIRES_AT = (new DateTime('now'))->modify("+$expires seconds");
        if( property_exists($rest_response, 'refresh_token') ) {
            $expires                                    = (0 + $rest_response->{'refresh_expires_in'}) -3;
            self::$KEYCLOAKAPI_REFRESH_TOKEN_VALUE      = $rest_response->{'refresh_token'};
            self::$KEYCLOAKAPI_REFRESH_TOKEN_EXPIRES_AT = (new DateTime('now'))->modify("+$expires seconds");

        }
        return self::$KEYCLOAKAPI_ACCESS_TOKEN_VALUE;
    }

    /**
     * Private utility function to obtain an user access token.
     * @return mixed in case of an error returns an array,
     *               in case of success it returns a string
     */
    private static function keycloackGetToken($p_aUserData) {
        if( !self::keycloackCheckParameters() ) {
            return array( 'ok'      => FALSE,
                          'message' => 'KEYCLOAK PARAMETERS NOT SET');
        }

        $p_strUserRealm    = NULL;
        $p_strUserName     = NULL;
        $p_strUserPassword = NULL;
        $p_strClientId     = NULL;
        $p_strClientSecret = NULL;

        if(isset($p_aUserData['realm'])) {
            $p_strUserRealm = $p_aUserData['realm'];
        }

        if(isset($p_aUserData['username'])) {
            $p_strUserName = $p_aUserData['username'];
        }

        if(isset($p_aUserData['password'])) {
            $p_strUserPassword = $p_aUserData['password'];
        }

        if(isset($p_aUserData['clientid'])) {
            $p_strClientId = $p_aUserData['clientid'];
        }

        if(isset($p_aUserData['clientsecret'])) {
            $p_strClientSecret = $p_aUserData['clientsecret'];
        }

        if( (empty($p_strUserRealm)    || trim($p_strUserRealm)    == '') &&
            (empty($p_strUserName)     || trim($p_strUserName)     == '') &&
            (empty($p_strUserPassword) || trim($p_strUserPassword) == '') &&
            (empty($p_strClientId)     || trim($p_strClientId)     == '') &&
            (empty($p_strClientSecret) || trim($p_strClientSecret) == '')
        )
        {
            //No valid way to identify a user was given.
            return array( 'ok'      => FALSE,
                          'message' => 'NO REALM/USERNAME/PASSWORD/CLIENTID/CLIENTSECRET PROVIDED');
        }

        $keycloak_url      = self::$KEYCLOAKAPI_URL;
        $adresse           = "$keycloak_url/realms/$p_strUserRealm/protocol/openid-connect/token";
        $methode           = "POST";
        $daten             = http_build_query(array(
                'grant_type'    => 'password',
                'username'      => $p_strUserName,
                'password'      => $p_strUserPassword,
                'client_id'     => $p_strClientId,
                'client_secret' => $p_strClientSecret
        ));
        $content_type = config('constants.keycloakSync.CONTENTTYPE_WWW_FORM_URLENCODED');

        $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, NULL);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )

        {
            return array( 'ok'       => FALSE,
                'message'  => 'REST CALL ERROR',
                'response' => $rest_response );
        }


        $rest_response = json_decode($rest_response['response']);

        if( !property_exists($rest_response, 'access_token') )
        {
            $error = 'NO ACCESS TOKEN IN RESPONSE';
            if( is_object( $rest_response ) && property_exists($rest_response, 'error') && property_exists($rest_response, 'error_description') )
            {
                $error = $rest_response->{'error_description'};
            }
            return array( 'ok'      => FALSE,
                'message' => $error);
        }

        return $rest_response;
    }

    /**
     * Private utility function to register a user.
     * @param $kc_username  string a username.
     * @param $kc_password  string a password.
     * @param $kc_firstname string a firstname.
     * @param $kc_lastname  string a lastname.
     * @param $kc_email     string an email.
     * @param $kc_mioid     integer a mioid.
     * @param $kc_adrid     integer a adrid.
     * @return array
     */
    private static function keycloakRegister($kc_username,
                                             $kc_password,
                                             $kc_firstname,
                                             $kc_lastname,
                                             $kc_email,
                                             $kc_dbid,
                                             $kc_mioid,
                                             $kc_adrid,
                                             $kc_phone,
                                             $kc_regtime )
    {
        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }

        $keycloak_url  = self::$KEYCLOAKAPI_URL;
        $adresse       = $keycloak_url . "/admin/realms/".self::$KEYCLOAKAPI_REALM."/users";
        $methode       = "POST";
        $content_type  = config('constants.keycloakSync.CONTENTTYPE_JSON');
        $daten         = json_encode(array(
            'username'  => $kc_username,
            'firstName' => $kc_firstname,
            'lastName'  => $kc_lastname,
            'email'     => $kc_email,
            'enabled'   => true,
            'attributes'=> array(
                self::$KEYCLOAKAPI_ATTRIBUTE_DBID  => $kc_dbid,
                self::$KEYCLOAKAPI_ATTRIBUTE_MIOID => $kc_mioid,
                self::$KEYCLOAKAPI_ATTRIBUTE_ADRID => $kc_adrid,
                self::$KEYCLOAKAPI_ATTRIBUTE_PHONE => $kc_phone,
                self::$KEYCLOAKAPI_ATTRIBUTE_UPDDA => $kc_regtime,
            )
        ));
        $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'REST CALL ERROR',
                          'response' => $rest_response );
        }

        if( !isset($rest_response['headers']['location']) )
        {
            $error = 'NO LOCATION';
            if( isset($rest_response['response']) )
            {
                $rest_response = json_decode($rest_response['response']);
                if( property_exists($rest_response, 'errorMessage') )
                {
                    $error = $rest_response->{'errorMessage'};
                }
            }
            else
            {
                $error .= ' AND NO RESPONSE';
            }
            return array( 'ok'      => FALSE,
                          'message' => $error);
        }

        $location = $rest_response['headers']['location'][0];
        $id       = preg_split("/\//", $location);
        $id       = $id[sizeof($id) - 1];

        if($kc_password !== NULL) {
            // Password is optional
            $adresse  = $keycloak_url . "/admin/realms/".self::$KEYCLOAKAPI_REALM."/users/" . $id . "/reset-password";
            $daten    = json_encode(array(
                'type'     => 'password',
                'value'    => $kc_password,
                'temporary'=> false
            ));
            $methode  = "PUT";
            $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, $access_token);

            if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
            {
                return array( 'ok'       => FALSE,
                              'message'  => 'REST CALL ERROR',
                              'response' => $rest_response );
            }

        }

        return array( 'ok'      => TRUE,
                      'message' => '',
                      'userid'  => $id);

    }

    private static function keycloakCreateRealmRole($p_aRoleData) {
        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }
        $keycloak_url  = self::$KEYCLOAKAPI_URL;
        $realm         = urlencode(self::$KEYCLOAKAPI_REALM);
        $adresse       = "$keycloak_url/admin/realms/$realm/roles";
        $daten         = json_encode($p_aRoleData);
        $content_type  = config('constants.keycloakSync.CONTENTTYPE_JSON');
        $methode       = "POST";
        $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'REST CALL ERROR',
                          'response' => $rest_response );
        }

        $ok                       = $rest_response['status'] == 201;
        $rest_response['ok']      = $ok;
        $rest_response['message'] = $ok?'':'THERE WAS AN ERROR TRYING TO CREATE THE REALM ROLE';
        return $rest_response;

    }

    private static function keycloakCreateClientRole($p_aRoleData) {
        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }
        $client_response = self::keycloakGetClientId(urlencode(self::$KEYCLOAKAPI_USER_CLIENT_ID));
        if(!$client_response['ok']) {
            return $client_response;
        }
        $keycloak_url  = self::$KEYCLOAKAPI_URL;
        $realm         = urlencode(self::$KEYCLOAKAPI_REALM);
        $client_id     = urlencode(json_decode($client_response['response'])[0]->{'id'});
        $adresse       = "$keycloak_url/admin/realms/$realm/clients/$client_id/roles";
        $daten         = json_encode($p_aRoleData);
        $content_type  = config('constants.keycloakSync.CONTENTTYPE_JSON');
        $methode       = "POST";
        $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'REST CALL ERROR',
                          'response' => $rest_response );
        }

        $ok                       = $rest_response['status'] == 200 || $rest_response['status'] == 201 || $rest_response['status'] == 204;
        $rest_response['ok']      = $ok;
        $rest_response['message'] = $ok?'':'THERE WAS AN ERROR TRYING TO CREATE THE CLIENT ROLE';
        return $rest_response;
    }

    private static function keycloakUserSetRealmRole($p_aUserData) {
        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }
        if(!isset($p_aUserData['id'])) {
            return array( 'ok'      => FALSE,
                          'message' => 'NO USERID PROVIDED');
        }
        if(!isset($p_aUserData['rolename'])) {
            return array( 'ok'      => FALSE,
                          'message' => 'NO ROLENAME PROVIDED');
        }
        $userid   = $p_aUserData['id'];
        $rolename = $p_aUserData['rolename'];
        $response = self::keycloakGetRealmRoleId($rolename);
        if($response['ok'] === FALSE) {
            return array( 'ok'      => $response['ok'],
                          'message' => $response['message']);
        }
        switch($response['status']) {
            case 200:
                $roleid = json_decode($response['response'])->{'id'};
                break;
            case 404:
                $response = self::keycloakCreateRealmRole(array('name' => $rolename));
                if($response['ok'] === FALSE) {
                    return array( 'ok'      => $response['ok'],
                                  'message' => $response['message']);
                }
                switch($response['status']) {
                    case 201:
                        $response = self::keycloakGetRealmRoleId($rolename);
                        if($response['ok'] === FALSE) {
                            return array( 'ok'      => $response['ok'],
                                          'message' => $response['message']);
                        }
                        $roleid = json_decode($response['response'])->{'id'};
                        break;
                }
                break;
        }

        $keycloak_url  = self::$KEYCLOAKAPI_URL;
        $realm         = urlencode(self::$KEYCLOAKAPI_REALM);
        $adresse       = "$keycloak_url/admin/realms/$realm/users/$userid/role-mappings/realm";
        $daten         = json_encode(array(array('id' => $roleid, 'name' => $rolename)));
        $content_type  = config('constants.keycloakSync.CONTENTTYPE_JSON');
        $methode       = "POST";
        $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'REST CALL ERROR',
                          'response' => $rest_response );
        }

        $ok            = $rest_response['status'] == 204;
        $rest_response['ok']      = $ok;
        $rest_response['message'] = $ok?'THE ROLE WAS ASSIGNED':'THERE WAS AN ERROR TRYING TO ASSIGN A ROLE';
        return $rest_response;
    }


    private static function keycloakUserGetRealmRoles($p_aUserData) {
        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }
        if(!isset($p_aUserData['id'])) {
            return array( 'ok'      => FALSE,
                          'message' => 'NO USERID PROVIDED');
        }
        $userid        = $p_aUserData['id'];
        $keycloak_url  = self::$KEYCLOAKAPI_URL;
        $realm         = urlencode(self::$KEYCLOAKAPI_REALM);
        $adresse       = "$keycloak_url/admin/realms/$realm/users/$userid/role-mappings/realm";
        $daten         = NULL;
        $content_type  = NULL;
        $methode       = "GET";
        $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'REST CALL ERROR',
                          'response' => $rest_response );
        }

        $ok                       = $rest_response['status'] == 200 || $rest_response['status'] == 404;
        $rest_response['ok']      = $ok;
        $rest_response['message'] = $ok?'':'THERE WAS AN ERROR TRYING TO GET THE REALM ROLES';
        return $rest_response;
    }

    private static function keycloakUserDeleteRealmRoles($p_aUserData) {
        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }
        if(!isset($p_aUserData['id'])) {
            return array( 'ok'      => FALSE,
                          'message' => 'NO USERID PROVIDED');
        }
        if(!isset($p_aUserData['roles'])) {
            return array( 'ok'      => FALSE,
                          'message' => 'NO ROLES PROVIDED');
        }
        $userid        = $p_aUserData['id'];
        $roles         = $p_aUserData['roles'];
        $keycloak_url  = self::$KEYCLOAKAPI_URL;
        $realm         = urlencode(self::$KEYCLOAKAPI_REALM);
        $adresse       = "$keycloak_url/admin/realms/$realm/users/$userid/role-mappings/realm";
        $daten         = json_encode($roles);
        $content_type  = config('constants.keycloakSync.CONTENTTYPE_JSON');
        $methode       = "DELETE";
        $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                'message'  => 'REST CALL ERROR',
                'response' => $rest_response );
        }

        $ok                       = $rest_response['status'] == 204;
        $rest_response['ok']      = $ok;
        $rest_response['message'] = $ok?'':'THERE WAS AN ERROR TRYING TO DELETE THE REALM ROLES';
        return $rest_response;
    }

    private static function keycloakUserSetClientRole($p_aUserData) {
         $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }
        if(!isset($p_aUserData['id'])) {
            return array( 'ok'      => FALSE,
                'message' => 'NO USERID PROVIDED');
        }
        if(!isset($p_aUserData['rolename'])) {
            return array( 'ok'      => FALSE,
                'message' => 'NO ROLENAME PROVIDED');
        }
        $client_response = self::keycloakGetClientId(urlencode(self::$KEYCLOAKAPI_USER_CLIENT_ID));
        if(!$client_response['ok']) {
            return $client_response;
        }
        $userid   = $p_aUserData['id'];
        $rolename = $p_aUserData['rolename'];
        $response = self::keycloakGetClientRoleId($rolename);
        if($response['ok'] === FALSE) {
            return array( 'ok'      => $response['ok'],
                'message' => $response['message']);
        }
        switch($response['status']) {
            case 200:
                $roleid = json_decode($response['response'])->{'id'};
                break;
            case 404:
                $response = self::keycloakCreateClientRole(array('name' => $rolename));
                if($response['ok'] === FALSE) {
                    return array( 'ok'      => $response['ok'],
                        'message' => $response['message']);
                }
                switch($response['status']) {
                    case 201:
                        $response = self::keycloakGetClientRoleId($rolename);
                        if($response['ok'] === FALSE) {
                            return array( 'ok'      => $response['ok'],
                                'message' => $response['message']);
                        }
                        $roleid = json_decode($response['response'])->{'id'};
                        break;
                }
                break;
        }

        $keycloak_url  = self::$KEYCLOAKAPI_URL;
        $realm         = urlencode(self::$KEYCLOAKAPI_REALM);
        $client_id     = urlencode(json_decode($client_response['response'])[0]->{'id'});
        $adresse       = "$keycloak_url/admin/realms/$realm/users/$userid/role-mappings/clients/$client_id";
        $daten         = json_encode(array(array('id' => $roleid, 'name' => $rolename)));
        $content_type  = config('constants.keycloakSync.CONTENTTYPE_JSON');
        $methode       = "POST";
        $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                'message'  => 'REST CALL ERROR',
                'response' => $rest_response );
        }

        $ok            = $rest_response['status'] == 204;
        $rest_response['ok']      = $ok;
        $rest_response['message'] = $ok?'THE ROLE WAS ASSIGNED':'THERE WAS AN ERROR TRYING TO ASSIGN A ROLE';
        return $rest_response;
    }

    private static function keycloakUserDeleteClientRole($p_aUserData) {
        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }
        if(!isset($p_aUserData['id'])) {
            return array( 'ok'      => FALSE,
                          'message' => 'NO USERID PROVIDED');
        }
        if(!isset($p_aUserData['rolename'])) {
            return array( 'ok'      => FALSE,
                          'message' => 'NO ROLENAME PROVIDED');
        }
        $client_response = self::keycloakGetClientId(urlencode(self::$KEYCLOAKAPI_USER_CLIENT_ID));
        if(!$client_response['ok']) {
            return $client_response;
        }
        $userid   = $p_aUserData['id'];
        $rolename = $p_aUserData['rolename'];
        $response = self::keycloakGetClientRoleId($rolename);
        if($response['ok'] === FALSE) {
            return array( 'ok'      => $response['ok'],
                          'message' => $response['message']);
        }
        switch($response['status']) {
            case 200:
                $roleid = json_decode($response['response'])->{'id'};
                break;
            case 404:
                return array( 'ok'      => $response['ok'],
                              'message' => $response['message']);
                break;
        }

        $keycloak_url  = self::$KEYCLOAKAPI_URL;
        $realm         = urlencode(self::$KEYCLOAKAPI_REALM);
        $client_id     = urlencode(json_decode($client_response['response'])[0]->{'id'});
        $adresse       = "$keycloak_url/admin/realms/$realm/users/$userid/role-mappings/clients/$client_id";
        $daten         = json_encode(array(array('id' => $roleid, 'name' => $rolename)));
        $content_type  = config('constants.keycloakSync.CONTENTTYPE_JSON');
        $methode       = "DELETE";
        $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                'message'  => 'REST CALL ERROR',
                'response' => $rest_response );
        }

        $ok            = $rest_response['status'] == 204;
        $rest_response['ok']      = $ok;
        $rest_response['message'] = $ok?'THE ROLE WAS REMOVED':'THERE WAS AN ERROR TRYING TO REMOVE A ROLE';
        return $rest_response;
    }

    public static function keycloakGetRealmRoleId($rolename) {
        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }
        $keycloak_url  = self::$KEYCLOAKAPI_URL;
        $realm         = urlencode(self::$KEYCLOAKAPI_REALM);
        $rolename      = urlencode($rolename);
        $adresse       = "$keycloak_url/admin/realms/$realm/roles/$rolename";
        $daten         = NULL;
        $content_type  = NULL;
        $methode       = "GET";
        $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'REST CALL ERROR',
                          'response' => $rest_response );
        }

        $ok                       = $rest_response['status'] == 200 || $rest_response['status'] == 404;
        $rest_response['ok']      = $ok;
        $rest_response['message'] = $ok?'':'THERE WAS AN ERROR TRYING TO GET THE REALM ROLE ID';
        return $rest_response;
    }

    private static function keycloakGetClientRoleId($rolename) {
        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }
        $client_response = self::keycloakGetClientId(urlencode(self::$KEYCLOAKAPI_USER_CLIENT_ID));
        if(!$client_response['ok']) {
            return $client_response;
        }
        $keycloak_url  = self::$KEYCLOAKAPI_URL;
        $realm         = urlencode(self::$KEYCLOAKAPI_REALM);
        $client_id     = urlencode(json_decode($client_response['response'])[0]->{'id'});
        $rolename      = urlencode($rolename);
        $adresse       = "$keycloak_url/admin/realms/$realm/clients/$client_id/roles/$rolename";
        $daten         = NULL;
        $content_type  = NULL;
        $methode       = "GET";
        $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'REST CALL ERROR TRYING TO ASSIGN CLIENT ROLE',
                          'response' => $rest_response );
        }

        $ok                       = $rest_response['status'] == 200 || $rest_response['status'] == 201 || $rest_response['status'] == 204;
        $rest_response['ok']      = $ok;
        $rest_response['message'] = $ok?'':'THERE WAS AN ERROR TRYING TO GET THE CLIENT ROLE ID';
        return $rest_response;
    }

    private static function keycloakGetClientId($clientname) {
        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }
        $keycloak_url  = self::$KEYCLOAKAPI_URL;
        $realm         = urlencode(self::$KEYCLOAKAPI_REALM);
        $clientname    = urlencode($clientname);
        $adresse       = "$keycloak_url/admin/realms/$realm/clients?clientId=$clientname";
        $daten         = NULL;
        $content_type  = NULL;
        $methode       = 'GET';
        $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'REST CALL ERROR TRYING TO GET CLIENT DATA',
                          'response' => $rest_response );
        }

        if( empty($rest_response['response']) ) {
            return array( 'ok'       => FALSE,
                          'message'  => 'REST CALL ERROR TRYING TO GET CLIENT ID: Response is empty, client may be lacking view-clients role.',
                          'response' => $rest_response );

        }

        $ok                       = $rest_response['status'] == 200 || $rest_response['status'] == 201 || $rest_response['status'] == 204;
        $rest_response['ok']      = $ok;
        $rest_response['message'] = $ok?'':'THERE WAS AN ERROR TRYING TO GET THE CLIENT ID';
        return $rest_response;
    }

    private static function keycloakSendUpdateAccountEmail($userid) {
        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }
        $keycloak_url  = self::$KEYCLOAKAPI_URL;
        $realm         = urlencode(self::$KEYCLOAKAPI_REALM);
        $adresse       = "$keycloak_url/admin/realms/$realm/users/$userid/execute-actions-email";
        $daten         = json_encode(array('Verify Email'));
        $content_type  = config('constants.keycloakSync.CONTENTTYPE_JSON');
        $methode       = 'PUT';
        $rest_response = self::rest_call($adresse, $methode, $daten, $content_type, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'REST CALL ERROR',
                          'response' => $rest_response );
        }

        $ok                       = $rest_response['status'] == 200 || $rest_response['status'] == 201 || $rest_response['status'] == 204;
        $rest_response['ok']      = $ok;
        $rest_response['message'] = $ok?'':'THERE WAS AN ERROR TRYING TO SEND UPDATE ACCOUNT EMAIL';
        return $rest_response;
    }


    /**
     * Private utility function that returns user info.
     * @param $p_aUserData array with the following keys/data for the user to find.
     *                           id is optional, but if you know it, this supersedes username/email
     *                           username is optional.
     *                           email is optional.
     * @return array
     */
    private static function keycloakUserStatus($p_aUserData)
    {
        $p_strUserName          = NULL;
        $p_strUserID            = NULL;
        $p_strEmail             = NULL;
        $p_strFirst             = NULL;
        $p_strMax               = NULL;
        $p_strQ                 = NULL;

        if(isset($p_aUserData['username'])) {
            $p_strUserName = $p_aUserData['username'];
        }

        if(isset($p_aUserData['id'])) {
            $p_strUserID = $p_aUserData['id'];
        }

        if(isset($p_aUserData['email'])) {
            $p_strEmail = $p_aUserData['email'];
        }

        if(isset($p_aUserData['first'])) {
            $p_strFirst = $p_aUserData['first'];
        }

        if(isset($p_aUserData['max'])) {
            $p_strMax = $p_aUserData['max'];
        }

        if(isset($p_aUserData['q'])) {
            $p_strQ = $p_aUserData['q'];
        }

//        if( (empty($p_strUserName) || trim($p_strUserName) == '') &&
//            (empty($p_strUserID)   || trim($p_strUserID)   == '') &&
//            (empty($p_strEmail)    || trim($p_strEmail)    == '')
//          )
//        {
//            //No valid way to identify a user was given.
//            return array( 'ok'       => FALSE,
//                          'message'  => 'NO ID/USERNAME/EMAIL PROVIDED',
//                          'response' => '');
//        }

        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }

        /// KEYCLOAK API 10.0
        ////////////////////////////////////////////////////////////
        /// briefRepresentation
        /// email
        /// first
        /// firstName
        /// lastName
        /// max
        /// search
        /// username
        ////////////////////////////////////////////////////////////
        /// Queries tested and working
        ///"/users?max=1
        ///"/users?briefRepresentation=false
        ///"/users?email=".urlencode('jhenriquez@comciencia.cl')
        ///"/users?lastName=".urlencode('Seitz')
        ///"/users?search=".urlencode('@comciencia.cl')
        ///"/users?username=".urlencode('jhenriquez@comciencia.cl')
        ////////////////////////////////////////////////////////////

        $keycloak_url = self::$KEYCLOAKAPI_URL;
        $realm        = urlencode(self::$KEYCLOAKAPI_REALM);
        if( $p_strUserID === NULL ) {
            $querystring = '';
            if (!empty($p_strUserName) && trim($p_strUserName) !== '') {
                $querystring .= "?username=" . urlencode($p_strUserName);
            }
            if (!empty($p_strEmail)    && trim($p_strEmail) !== '') {
                $querystring .= ($querystring === ''?'?':'&') . "email=" . urlencode($p_strEmail);
            }
            if (!empty($p_strFirst) ) {
                $querystring .= ($querystring === ''?'?':'&') . "first=" . urlencode($p_strFirst);
            }
            if (!empty($p_strMax) ) {
                $querystring .= ($querystring === ''?'?':'&') . "max=" . urlencode($p_strMax);
            }
            if (!empty($p_strQ) ) {
                $querystring .= ($querystring === ''?'?':'&') . "q=" . urlencode($p_strQ);
            }
            $adresse = "$keycloak_url/admin/realms/$realm/users$querystring";
        }
        else{
            $adresse = "$keycloak_url/admin/realms/$realm/users/$p_strUserID";
        }

        //Log::debug( "Adresse $adresse" );

        $rest_response = self::rest_call($adresse, 'GET', NULL, NULL, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'REST CALL ERROR',
                          'response' => $rest_response );
        }

        $error              = '';
        $temp_rest_response = json_decode($rest_response['response']);

        if( $rest_response['status'] != 200 &&
            is_object( $temp_rest_response ) ) {
            if( property_exists($temp_rest_response, 'errorMessage') ) {
                $error = $temp_rest_response->{'errorMessage'};
                return array( 'ok'       => FALSE,
                              'message'  => $error,
                              'response' => $rest_response );
            }
            if( property_exists($temp_rest_response, 'error') ) {
                $error = $temp_rest_response->{'error'};
                return array( 'ok'       => FALSE,
                              'message'  => $error,
                              'response' => $rest_response );
            }
        }

        if( $rest_response['status'] == 200 &&
            is_object( $temp_rest_response ) ) {
            //THERE WAS NO ERROR, BUT AN OBJECT WAS RETURNED
            //THIS IS THE CASE WHEN YOU SEARCH BY ID
            //WE WRAP IT AS AN ARRAY
            $temp_rest_response = array($temp_rest_response);
        }

        if( $rest_response['status'] == 200 &&
            is_array( $temp_rest_response ) ) {
            //RESPONSE IS AN ARRAY, TENDING TO BE OK;
            //CHECK IF KEYS REQUIRED ARE PRESENT
            if( count($temp_rest_response ) == 0) {
                return array( 'ok'       => FALSE,
                              'message'  => 'USER NOT FOUND',
                              'response' => $rest_response );
            }
            if( count($temp_rest_response) > 1 ) {
                return array( 'ok'       => FALSE,
                              'message'  => 'TOO MANY USERS FOUND',
                              'response' => $rest_response );
            }
            if( !is_object($temp_rest_response[0]) ) {
                return array( 'ok'       => FALSE,
                              'message'  => 'RESPONSE ARRAY DOES NOT CONTAIN AN OBJECT',
                              'response' => $rest_response );
            }

            $aNames = array(
                'access',
                'attributes',
                'clientConsents',
                'clientRoles',
                'createdTimestamp',
                'credentials',
                'disableableCredentialTypes',
                'email',
                'emailVerified',
                'enabled',
                'federatedIdentities',
                'federationLink',
                'firstName',
                'groups',
                'id',
                'lastName',
                'notBefore',
                'origin',
                'realmRoles',
                'requiredActions',
                'self',
                'serviceAccountClientId',
                'username');

            $rest_response = array();
            foreach($aNames as $sName) {
                if( !property_exists($temp_rest_response[0],$sName) ) continue;
                if( $sName === 'attributes' ) {
                    if( is_object($temp_rest_response[0]->{$sName}) ) {
                        $rest_response[$sName] = array();
                        foreach($temp_rest_response[0]->{$sName} as $attr_name => $attr_value) {
                            if (is_array($attr_value)) {
                                $rest_response[$sName][$attr_name] = array_shift($attr_value);
                                continue;
                            }
                            $rest_response[$sName][$attr_name] = $attr_value;
                        }
                    }
                    continue;
                }
                $rest_response[$sName] = $temp_rest_response[0]->{$sName};
            }
        }
        else{
            //MAYBE PLAIN/HTML OR PLAIN/TEXT ??
            $error = 'RESPONSE IS NOT AN OBJECT NOR AN ARRAY';
        }

        return array( 'ok'       => empty($error),
                      'message'  => empty($error)?'USER FOUND':$error,
                      'response' => $rest_response );

    }

    /**
     * Obtain number of users in the given realm.
     *
     * @return array|int|mixed array on error, number on success
     */
   public static function keycloakUsersCount()
    {
        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }

        $keycloak_url = self::$KEYCLOAKAPI_URL;
        $realm        = urlencode(self::$KEYCLOAKAPI_REALM);

        $adresse = "$keycloak_url/admin/realms/$realm/users/count";

        //Log::debug( "Adresse $adresse" );

        $rest_response = self::rest_call($adresse, 'GET', NULL, NULL, $access_token);

        if( $rest_response["status"] == 200 ) {
            return $rest_response["response"];
        }
        else {
            return array();
        }
    }

   public static function keycloakUserGetClientRole($aUserData)
    {
        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }

        $client_response = self::keycloakGetClientId(urlencode(self::$KEYCLOAKAPI_USER_CLIENT_ID));
        if(!$client_response['ok']) {
            return $client_response;
        }
        $client_id     = urlencode(json_decode($client_response['response'])[0]->{'id'});

        $keycloak_url = self::$KEYCLOAKAPI_URL;
        $realm        = urlencode(self::$KEYCLOAKAPI_REALM);
        $adresse = "$keycloak_url/admin/realms/$realm/users/" . $aUserData["id"] . "/role-mappings/clients/" . $client_id;

        Log::debug( "Adresse $adresse" );

        $rest_response = self::rest_call($adresse, 'GET', NULL, NULL, $access_token);

        //Log::debug( "keycloakUserGetClientRole response " . print_r($rest_response, true) );

        $aResp = json_decode($rest_response["response"]);
        if( $rest_response["status"] == 200 && isset($aResp[0]) ) {
            return $aResp[0]->name;
        }
        else {
            return array();
        }
    }


    /**
     * Private utility function to update User info
     * @param $p_aUserData array    mandatory, data to update with at least one of 'id','username' or 'email'
     *                              and the following optional keys and data:
     *                              'access', 'attributes', 'clientConsents', 'clientRoles',
     *                              'createdTimestamp', 'credentials', 'disableableCredentialTypes'
     *                              'emailVerified', 'enabled', 'federatedIdentities',
     *                              'federationLink', 'firstName', 'groups', 'lastName',
     *                              'notBefore', 'origin', 'realmRoles', 'requiredActions',
     *                              'self', 'serviceAccountClientId'.
     * @return array
     */
    private static function keycloakUserUpdate($p_aUserData) {
        if( empty($p_aUserData) || !is_array($p_aUserData) || count($p_aUserData) == 0 )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'NO USERDATA PROVIDED',
                          'response' => '' );
        }

        $p_strUserName = NULL;
        $p_strUserID   = NULL;
        $p_strEmail    = NULL;

        if(isset($p_aUserData['username'])) {
            $p_strUserName = $p_aUserData['username'];
        }

        if(isset($p_aUserData['id'])) {
            $p_strUserID = $p_aUserData['id'];
        }

        if(isset($p_aUserData['email'])) {
            $p_strEmail = $p_aUserData['email'];
        }

        if( (empty($p_strUserName) || trim($p_strUserName) == '') &&
            (empty($p_strUserID)   || trim($p_strUserID)   == '') &&
            (empty($p_strEmail)    || trim($p_strEmail)    == '')
        )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'NO USERID/USERNAME/EMAIL PROVIDED',
                          'response' => '');
        }

        $aNames = array(
            'access',
            'attributes',
            'clientConsents',
            'clientRoles',
            'createdTimestamp',
            'credentials',
            'disableableCredentialTypes',
            'email',
            'emailVerified',
            'enabled',
            'federatedIdentities',
            'federationLink',
            'firstName',
            'groups',
            'id',
            'lastName',
            'notBefore',
            'origin',
            'realmRoles',
            'requiredActions',
            'self',
            'serviceAccountClientId',
            'username');

        $p_aUpdatedData = array();
        foreach($aNames as $sName) {
            if(!isset($p_aUserData[$sName])) continue;
            $p_aUpdatedData[$sName] = $p_aUserData[$sName];
        }
        if( empty($p_aUpdatedData) )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'NO USERDATA TO UPDATE',
                          'response' => '');
        }

        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }

        if($p_strUserID === NULL) {
            $aUserToQuery = array();
            if($p_strUserName !== NULL) {
                $aUserToQuery['username'] = $p_strUserName;
            }
            if($p_strEmail !== NULL) {
                $aUserToQuery['email'] = $p_strEmail;
            }
            $return = self::keycloakUserStatus($aUserToQuery);
            if(!isset($return['ok'])) {
                return array( 'ok'       => FALSE,
                              'message'  => 'SOMETHING WENT WRONG',
                              'response' => '');
            }
            if($return['ok'] === FALSE ) {
                return $return;
            }
            if(!isset($return['response'])) {
                return array( 'ok'       => FALSE,
                              'message'  => 'SOMETHING WENT WRONG',
                              'response' => '');
            }
            if(!isset($return['response']['id'])) {
                return array( 'ok'       => FALSE,
                              'message'  => 'SOMETHING WENT WRONG',
                              'response' => '');
            }
            $id = $return['response']['id'];
        }
        else{
            $id = $p_strUserID;
        }
        $keycloak_url   = self::$KEYCLOAKAPI_URL;
        $realm          = urlencode(self::$KEYCLOAKAPI_REALM);
        $content_type   = config('constants.keycloakSync.CONTENTTYPE_JSON');
        $daten          = json_encode( $p_aUpdatedData );
        $adresse        = "$keycloak_url/admin/realms/$realm/users/$id";
        $rest_response  = self::rest_call($adresse, 'PUT', $daten, $content_type, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'REST CALL ERROR',
                          'response' => $rest_response );
        }

        if( !is_string($rest_response['response']) || $rest_response['response'] !== '' ) {
            return array( 'ok'       => FALSE,
                          'message'  => 'SOMETHING WENT WRONG',
                          'response' => $rest_response);
        }

        return array( 'ok'       => TRUE,
                      'message'  => '',
                      'response' => $rest_response);

    }

    /**
     * Private utility function to delete an user.
     * @param $p_strUserName string a user to delete.
     * @param $p_strEmail string optional, a mail for a user to delete.
     * @return array
     */
    private static function keycloakUserDelete($p_aUserData) {
        $p_strUserName = NULL;
        $p_strUserID   = NULL;
        $p_strEmail    = NULL;

        if(isset($p_aUserData['username'])) {
            $p_strUserName = $p_aUserData['username'];
        }

        if(isset($p_aUserData['id'])) {
            $p_strUserID = $p_aUserData['id'];
        }

        if(isset($p_aUserData['email'])) {
            $p_strEmail = $p_aUserData['email'];
        }

        if( (empty($p_strUserName) || trim($p_strUserName) == '') &&
            (empty($p_strUserID)   || trim($p_strUserID)   == '') &&
            (empty($p_strEmail)    || trim($p_strEmail)    == '')
        )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'NO USERID/USERNAME/EMAIL PROVIDED',
                          'response' => '');
        }

        $access_token = self::keycloackGetAdminToken();
        if( is_array($access_token) ) {
            return $access_token;
        }

        if($p_strUserID === NULL) {
            $aUserToQuery = array();
            if($p_strUserName !== NULL) {
                $aUserToQuery['username'] = $p_strUserName;
            }
            if($p_strEmail !== NULL) {
                $aUserToQuery['email'] = $p_strEmail;
            }
            $return = self::keycloakUserStatus($aUserToQuery);
            if(!isset($return['ok'])) {
                return array( 'ok'       => FALSE,
                              'message'  => 'SOMETHING WENT WRONG',
                              'response' => '');
            }
            if($return['ok'] === FALSE ) {
                return $return;
            }
            if(!isset($return['response'])) {
                return array( 'ok'       => FALSE,
                              'message'  => 'SOMETHING WENT WRONG',
                              'response' => '');
            }
            if(!isset($return['response']['id'])) {
                return array( 'ok'       => FALSE,
                              'message'  => 'SOMETHING WENT WRONG',
                              'response' => '');
            }
            $id = $return['response']['id'];
        }
        else{
            $id = $p_strUserID;
        }

        $keycloak_url  = self::$KEYCLOAKAPI_URL;
        $realm         = urlencode(self::$KEYCLOAKAPI_REALM);
        $adresse       = "$keycloak_url/admin/realms/$realm/users/$id";
        $rest_response = self::rest_call($adresse, 'DELETE', NULL, NULL, $access_token);

        if( $rest_response['status'] == 408 || $rest_response['status'] > 899 )
        {
            return array( 'ok'       => FALSE,
                          'message'  => 'REST CALL ERROR',
                          'response' => $rest_response );
        }

        if( !is_string($rest_response['response']) || $rest_response['response'] !== '' ) {
            return array( 'ok'       => FALSE,
                          'message'  => 'SOMETHING WENT WRONG',
                          'response' => $rest_response);
        }
        return array( 'ok'       => TRUE,
                      'message'  => '',
                      'response' => $rest_response);
    }

    //┌──────────────────────────────────────────────────────────────────────────┐
    //│                                                                          │
    //│                                                                          │
    //│           ########  ##     ## ########  ##       ####  ######            │
    //│           ##     ## ##     ## ##     ## ##        ##  ##    ##           │
    //│           ##     ## ##     ## ##     ## ##        ##  ##                 │
    //│           ########  ##     ## ########  ##        ##  ##                 │
    //│           ##        ##     ## ##     ## ##        ##  ##                 │
    //│           ##        ##     ## ##     ## ##        ##  ##    ##           │
    //│           ##         #######  ########  ######## ####  ######            │
    //│                                                                          │
    //│                                                                          │
    //└──────────────────────────────────────────────────────────────────────────┘

    /**
     * Public utility function to set Keycloak parameters.
     * @param $p_aParameters array of strings with the following keys/values
     *                             'Parameters' which is also an array, with the following keys
     *                             'KEYCLOAKAPI_REALM', 'KEYCLOAKAPI_URL', 'KEYCLOAKAPI_ADMIN_USERNAME',
     *                             'KEYCLOAKAPI_ADMIN_PASSWORD', 'KEYCLOAKAPI_ADMIN_CLIENT_ID',
     *                             'KEYCLOAKAPI_USER_CLIENT_SECRET'
     *
     *                             'Roles' which is also an array, with the following keys
     *                             'KEYCLOAKAPI_ATTRIBUTE_MIOID', 'KEYCLOAKAPI_ATTRIBUTE_ADRID'

     *                             'Attributes' which is also an array, with the following keys
     *                             'KEYCLOAKAPI_ATTRIBUTE_MIOID', 'KEYCLOAKAPI_ATTRIBUTE_ADRID'
     *
     * @return array
     */
    public static function setConfiguration($p_aConfiguration) {
        $p_aParameters = $p_aConfiguration['Parameters'];
        $p_aRoles      = $p_aConfiguration['Roles'];
        $p_aRoles_Desc = $p_aConfiguration['Roles-Descriptions'];
        $p_aAttributes = $p_aConfiguration['Attributes'];
        if( !isset($p_aParameters['KEYCLOAKAPI_REALM'])                         ||
              trim($p_aParameters['KEYCLOAKAPI_REALM'])                   == '' ||
            !isset($p_aParameters['KEYCLOAKAPI_URL'])                           ||
              trim($p_aParameters['KEYCLOAKAPI_URL'])                     == '' ||
            !isset($p_aParameters['KEYCLOAKAPI_ADMIN_USERNAME'])                ||
              trim($p_aParameters['KEYCLOAKAPI_ADMIN_USERNAME'])          == '' ||
            !isset($p_aParameters['KEYCLOAKAPI_ADMIN_PASSWORD'])                ||
              trim($p_aParameters['KEYCLOAKAPI_ADMIN_PASSWORD'])          == '' ||
            !isset($p_aParameters['KEYCLOAKAPI_ADMIN_CLIENT_ID'])               ||
              trim($p_aParameters['KEYCLOAKAPI_ADMIN_CLIENT_ID'])         == '' ||
            !isset($p_aParameters['KEYCLOAKAPI_USER_CLIENT_ID'])                ||
              trim($p_aParameters['KEYCLOAKAPI_USER_CLIENT_ID'])          == '' ||
            !isset($p_aParameters['KEYCLOAKAPI_USER_CLIENT_SECRET'])            ||
              trim($p_aParameters['KEYCLOAKAPI_USER_CLIENT_SECRET'])      == '' ||
            !isset($p_aRoles['KEYCLOAKAPI_REALM_ROLE_CLUB_MEMBERS'])            ||
              trim($p_aRoles['KEYCLOAKAPI_REALM_ROLE_CLUB_MEMBERS'])      == '' ||
            !isset($p_aRoles['KEYCLOAKAPI_REALM_ROLE_CLUB_ADMINS'])             ||
              trim($p_aRoles['KEYCLOAKAPI_REALM_ROLE_CLUB_ADMINS'])       == '' ||
            !isset($p_aRoles['KEYCLOAKAPI_REALM_ROLE_CLUB_GUESTS'])             ||
              trim($p_aRoles['KEYCLOAKAPI_REALM_ROLE_CLUB_GUESTS'])       == '' ||
            !isset($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_ADMIN'])                  ||
              trim($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_ADMIN'])            == '' ||
            !isset($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_USER'])                   ||
              trim($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_USER'])             == '' ||
            !isset($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_GUEST'])                  ||
              trim($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_GUEST'])            == '' ||
            !isset($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_FUNCTIONARY'])            ||
              trim($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_FUNCTIONARY'])      == '' ||
            !isset($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_SECRETARY'])              ||
              trim($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_SECRETARY'])        == '' ||
            !isset($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_EDITOR'])                 ||
              trim($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_EDITOR'])           == '' ||
            !isset($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_MEMBER_LIGHT'])           ||
              trim($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_MEMBER_LIGHT'])     == '' ||
            !isset($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_ADMIN_LIGHT'])            ||
              trim($p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_ADMIN_LIGHT'])      == '' ||
            !isset($p_aRoles_Desc['KEYCLOAKAPI_REALM_ROLE_CLUB_MEMBERS'])       ||
              trim($p_aRoles_Desc['KEYCLOAKAPI_REALM_ROLE_CLUB_MEMBERS']) == '' ||
            !isset($p_aRoles_Desc['KEYCLOAKAPI_REALM_ROLE_CLUB_ADMINS'])        ||
              trim($p_aRoles_Desc['KEYCLOAKAPI_REALM_ROLE_CLUB_ADMINS'])  == '' ||
            !isset($p_aRoles_Desc['KEYCLOAKAPI_REALM_ROLE_CLUB_GUESTS'])        ||
              trim($p_aRoles_Desc['KEYCLOAKAPI_REALM_ROLE_CLUB_GUESTS'])  == '' ||
            !isset($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_ADMIN'])             ||
              trim($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_ADMIN'])       == '' ||
            !isset($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_USER'])              ||
              trim($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_USER'])        == '' ||
            !isset($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_GUEST'])             ||
              trim($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_GUEST'])       == '' ||
            !isset($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_FUNCTIONARY'])       ||
              trim($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_FUNCTIONARY']) == '' ||
            !isset($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_SECRETARY'])         ||
              trim($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_SECRETARY'])   == '' ||
            !isset($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_EDITOR'])            ||
              trim($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_EDITOR'])      == '' ||
            !isset($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_MEMBER_LIGHT'])      ||
              trim($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_MEMBER_LIGHT']) == '' ||
            !isset($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_ADMIN_LIGHT'])       ||
              trim($p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_ADMIN_LIGHT']) == '' ||
            !isset($p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_DBID'])                ||
              trim($p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_DBID'])          == '' ||
            !isset($p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_MIOID'])               ||
              trim($p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_MIOID'])         == '' ||
            !isset($p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_ADRID'])               ||
              trim($p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_ADRID'])         == '' ||
            !isset($p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_PHONE'])               ||
              trim($p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_PHONE'])         == '' ||
            !isset($p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_UPDDA'])               ||
              trim($p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_UPDDA'])         == '') {

            return array( 'ok'      => FALSE,
                          'message' => 'MISSING KEYCLOAK PARAMETER/ROLE/ATTRIBUTE');
        }

        self::$KEYCLOAKAPI_REALM                         = $p_aParameters['KEYCLOAKAPI_REALM'];
        self::$KEYCLOAKAPI_URL                           = $p_aParameters['KEYCLOAKAPI_URL'];
        self::$KEYCLOAKAPI_ADMIN_USERNAME                = $p_aParameters['KEYCLOAKAPI_ADMIN_USERNAME'];
        self::$KEYCLOAKAPI_ADMIN_PASSWORD                = $p_aParameters['KEYCLOAKAPI_ADMIN_PASSWORD'];
        self::$KEYCLOAKAPI_ADMIN_CLIENT_ID               = $p_aParameters['KEYCLOAKAPI_ADMIN_CLIENT_ID'];
        self::$KEYCLOAKAPI_USER_CLIENT_ID                = $p_aParameters['KEYCLOAKAPI_USER_CLIENT_ID'];
        self::$KEYCLOAKAPI_USER_CLIENT_SECRET            = $p_aParameters['KEYCLOAKAPI_USER_CLIENT_SECRET'];

        self::$KEYCLOAKAPI_REALM_ROLE_CLUB_MEMBERS       = $p_aRoles['KEYCLOAKAPI_REALM_ROLE_CLUB_MEMBERS'];
        self::$KEYCLOAKAPI_REALM_ROLE_CLUB_ADMINS        = $p_aRoles['KEYCLOAKAPI_REALM_ROLE_CLUB_ADMINS'];
        self::$KEYCLOAKAPI_REALM_ROLE_CLUB_GUESTS        = $p_aRoles['KEYCLOAKAPI_REALM_ROLE_CLUB_GUESTS'];
        self::$KEYCLOAKAPI_CLIENT_ROLE_ADMIN             = $p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_ADMIN'];
        self::$KEYCLOAKAPI_CLIENT_ROLE_USER              = $p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_USER'];
        self::$KEYCLOAKAPI_CLIENT_ROLE_GUEST             = $p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_GUEST'];
        self::$KEYCLOAKAPI_CLIENT_ROLE_FUNCTIONARY       = $p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_FUNCTIONARY'];
        self::$KEYCLOAKAPI_CLIENT_ROLE_SECRETARY         = $p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_SECRETARY'];
        self::$KEYCLOAKAPI_CLIENT_ROLE_EDITOR            = $p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_EDITOR'];
        self::$KEYCLOAKAPI_CLIENT_ROLE_MEMBER_LIGHT      = $p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_MEMBER_LIGHT'];
        self::$KEYCLOAKAPI_CLIENT_ROLE_ADMIN_LIGHT       = $p_aRoles['KEYCLOAKAPI_CLIENT_ROLE_ADMIN_LIGHT'];

        self::$KEYCLOAKAPI_DESC_REALM_ROLE_CLUB_MEMBERS  = $p_aRoles_Desc['KEYCLOAKAPI_REALM_ROLE_CLUB_MEMBERS'];
        self::$KEYCLOAKAPI_DESC_REALM_ROLE_CLUB_ADMINS   = $p_aRoles_Desc['KEYCLOAKAPI_REALM_ROLE_CLUB_ADMINS'];
        self::$KEYCLOAKAPI_DESC_REALM_ROLE_CLUB_GUESTS   = $p_aRoles_Desc['KEYCLOAKAPI_REALM_ROLE_CLUB_GUESTS'];
        self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_ADMIN        = $p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_ADMIN'];
        self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_USER         = $p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_USER'];
        self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_GUEST        = $p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_GUEST'];
        self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_FUNCTIONARY  = $p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_FUNCTIONARY'];
        self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_SECRETARY    = $p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_SECRETARY'];
        self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_EDITOR       = $p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_EDITOR'];
        self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_MEMBER_LIGHT = $p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_MEMBER_LIGHT'];
        self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_ADMIN_LIGHT  = $p_aRoles_Desc['KEYCLOAKAPI_CLIENT_ROLE_ADMIN_LIGHT'];

        self::$KEYCLOAKAPI_ATTRIBUTE_DBID                = $p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_DBID'];
        self::$KEYCLOAKAPI_ATTRIBUTE_MIOID               = $p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_MIOID'];
        self::$KEYCLOAKAPI_ATTRIBUTE_ADRID               = $p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_ADRID'];
        self::$KEYCLOAKAPI_ATTRIBUTE_PHONE               = $p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_PHONE'];
        self::$KEYCLOAKAPI_ATTRIBUTE_UPDDA               = $p_aAttributes['KEYCLOAKAPI_ATTRIBUTE_UPDDA'];

        if( isset($p_aAttributes['KEYCLOAKAPI_CONNECTTIMEOUT']) &&
            is_numeric($p_aAttributes['KEYCLOAKAPI_CONNECTTIMEOUT'])) {
            self::$KEYCLOAKAPI_CONNECTTIMEOUT = $p_aAttributes['KEYCLOAKAPI_CONNECTTIMEOUT'];
        }

        if( isset($p_aAttributes['KEYCLOAKAPI_TIMEOUT']) &&
            is_numeric($p_aAttributes['KEYCLOAKAPI_TIMEOUT'])) {
            self::$KEYCLOAKAPI_TIMEOUT = $p_aAttributes['KEYCLOAKAPI_TIMEOUT'];
        }

        return array( 'ok'      => TRUE,
                      'message' => '');
    }

    /**
     * Public utility function to chek Keycloak parameters.
     *
     * @return boolean
     */
    public static function checkConfiguration() {
        return self::keycloackCheckParameters();
    }

    /**
     * Returns information for a given user. You must call the Keycloak::setConfiguration function first
     * @param $p_aUserData array with the following keys/data for the user to find.
     *                           id is optional, but if you know it, this supersedes username/email
     *                           username is optional.
     *                           email is optional.
     * @return
     *         multiple types
     *         integer  -3 in case username is empty
     *         string   in case keycloak api encounters an error
     *         boolean  FALSE in case of an error in rest api
     *         array    with the following keys and data
     *                  'id','createdTimestamp','username',
     *                  'enabled','emailVerified','firstName',
     *                  'lastName','email','attributes',
     *                  'disableableCredentialTypes',
     *                  'requiredActions','access'
     */
    public static function statusUser($p_aUserData) {
        $return = self::keycloakUserStatus($p_aUserData);
        if( !isset($return['ok']) || ( $return['ok'] !== TRUE && $return['ok'] !== FALSE ) ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        if( !isset($return['message']) ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        if( !isset($return['response']) ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        return $return;
    }

    /**
     * Enables the given user. You must call the Keycloak::setConfiguration function first
     * in order to set Keycloak parameters.
     * @param $p_aUserData array with the following keys/data for the user to ENABLE.
     *                           id is optional, but if you know it, this supersedes username/email
     *                           username is optional.
     *                           email is optional.
     * @return
     *         multiple types
     *         integer  -3 in case username is empty
     *         boolean  FALSE in case of user not found or an error in rest api
     *         boolean  TRUE  on success
     */
    public static function enableUser($p_aUserData) {
        $p_aUserData['enabled'] = TRUE;
        $return = Keycloak::keycloakUserUpdate($p_aUserData);
        if( !isset($return['ok']) || ( $return['ok'] !== TRUE && $return['ok'] !== FALSE ) ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        if( !isset($return['message']) ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        if( !isset($return['response']) ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        return $return;
    }

    /**
     * Disables the given User. You must call the Keycloak::setConfiguration function first
     * in order to set Keycloak parameters.
     * @param $p_aUserData array with the following keys/data for the user to DISABLE.
     *                           id is optional, but if you know it, this supersedes username/email
     *                           username is optional.
     *                           email is optional.
     * @return
     *         multiple types
     *         integer  -3 in case username is empty.
     *         boolean  FALSE in case of user not found or an error in rest api.
     *         boolean  TRUE  on success.
     */
    public static function disableUser($p_aUserData) {
        $p_aUserData['enabled'] = FALSE;
        $return = Keycloak::keycloakUserUpdate($p_aUserData);
        if( !isset($return['ok']) || ( $return['ok'] !== TRUE && $return['ok'] !== FALSE ) ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        if( !isset($return['message']) ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        if( !isset($return['response']) ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        return $return;
    }

    /**
     * Enables the given user. You must call the Keycloak::setConfiguration function first
     * in order to set Keycloak parameters.
     * @param $p_aUserData array with the following keys/data for the user to ENABLE.
     *                           id is optional, but if you know it, this supersedes username/email
     *                           username is optional.
     *                           email is optional.
     * @return
     *         multiple types
     *         integer  -3 in case username is empty
     *         boolean  FALSE in case of user not found or an error in rest api
     *         boolean  TRUE  on success
     */
    public static function updateUser($p_aUserData) {
        if( isset($p_aUserData['attributes']) ) {
            if(self::$KEYCLOAKAPI_TIME == NULL) {
                $time = (new DateTime('now'))->format('Y.m.d.H.i.s');;
            }
            else{
                $time = self::$KEYCLOAKAPI_TIME;
            }
            $p_aUserData['attributes'][self::$KEYCLOAKAPI_ATTRIBUTE_UPDDA] = $time;
        }
        $return = Keycloak::keycloakUserUpdate($p_aUserData);
        if( !isset($return['ok']) || ( $return['ok'] !== TRUE && $return['ok'] !== FALSE ) ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        if( !isset($return['message']) ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        if( !isset($return['response']) ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        return $return;
    }

    /**
     * Deletes the given User. You must call the Keycloak::setConfiguration function first
     * in order to set Keycloak parameters.
     * @param $p_aUserData array with the following keys/data for the user to DELETE.
     *                           id is optional, but if you know it, this supersedes username/email
     *                           username is optional.
     *                           email is optional.
     * @return
     *         multiple types
     *         integer  -3 in case username is empty.
     *         boolean  FALSE in case of user not found or an error in rest api.
     *         boolean  TRUE  on success.
     */
    public static function deleteUser($p_aUserData) {
        $return = Keycloak::keycloakUserDelete($p_aUserData);
        if( !isset($return['ok']) || $return['ok'] === FALSE ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        if( !isset($return['message']) ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        if( !isset($return['response']) ) {
            return config('constants.keycloakSync.KEYCLOAKAPI_ERROR');
        }
        return $return;
    }

  /**
   * @param $p_iDbId
   * @param $p_iKundeId
   * @param $p_aUserdata
   * @return array|false|string[]
   */
    public static function verifyUser($p_iDbId, $p_iKundeId, $p_aUserdata )
    {
      // Parameter validation
      if (!isset($p_aUserdata['search'])) {
        return FALSE;
      }
      if ($p_aUserdata['search'] != 'userid' && $p_aUserdata['search'] != 'email' && $p_aUserdata['search'] != 'username') {
        return FALSE;
      }
      if ($p_aUserdata['search'] == 'userid' && !isset($p_aUserdata['userid'])) {
        return FALSE;
      }
      if ($p_aUserdata['search'] == 'email' && !isset($p_aUserdata['email'])) {
        return FALSE;
      }
      if (!isset($p_aUserdata['adrid'])) {
        return FALSE;
      }

      $userid = (isset($p_aUserdata['userid']) ? $p_aUserdata['userid'] : '');
      $email = $p_aUserdata['email'];

      if ($p_aUserdata['search'] == 'userid') {
        $return = self::statusUser(array('id' => $userid));
      }
      elseif ($p_aUserdata['search'] == 'username') {
        $return = self::statusUser(array('username' => $p_aUserdata['username']));
      }
      else {
        $return = self::statusUser(array('username' => $email));
      }

      if (is_array($return) && isset($return['ok']) && $return['ok'] === config('constants.keycloakSync.KEYCLOAKAPI_SUCCESS')) {
        if (isset($return['response']['username']) && isset($return['response']['email']) &&
            $return['response']['email'] == $email) {
          //FOUND IT
          //CHECK FOR OTHER PROPERTIES: ATTRIBUTES AND ROLE
          if (!isset($return['response']['attributes']) || !is_array($return['response']['attributes'])) {
            //there was an error, the returned value does not have what we need
            return array('STATISTIC' => 'ERR-User-No-Attributes', 'RETURN' => 'KEYCLOAK attribute abschnitt leer oder nicht richtig eingestellt');
          }
          if (!isset($return['response']['attributes'][self::$KEYCLOAKAPI_ATTRIBUTE_DBID])) {
            //there was an error, the returned value does not have what we need
            return array('STATISTIC' => 'ERR-User-No-Attributes', 'RETURN' => sprintf('KEYCLOAK attribut %s fehlt', self::$KEYCLOAKAPI_ATTRIBUTE_DBID));
          }
          if (!isset($return['response']['attributes'][self::$KEYCLOAKAPI_ATTRIBUTE_MIOID])) {
            //there was an error, the returned value does not have what we need
            return array('STATISTIC' => 'ERR-User-No-Attributes', 'RETURN' => sprintf('KEYCLOAK attribut %s fehlt', self::$KEYCLOAKAPI_ATTRIBUTE_MIOID));
          }
          if (!isset($return['response']['attributes'][self::$KEYCLOAKAPI_ATTRIBUTE_ADRID])) {
            //there was an error, the returned value does not have what we need
            return array('STATISTIC' => 'ERR-User-No-Attributes', 'RETURN' => sprintf('KEYCLOAK attribut %s fehlt', self::$KEYCLOAKAPI_ATTRIBUTE_ADRID));
          }
          //Here the email was found, is it in the same club, is the same user
          if ($return['response']['attributes'][self::$KEYCLOAKAPI_ATTRIBUTE_DBID] != $p_iDbId) {
            //Not in the same club, this should be treated as an error??
            //Or a user not invited???
            //there was an error
            return array('STATISTIC' => 'ERR-User-Distinct-Dbid', 'RETURN' => $return['response']['attributes'][self::$KEYCLOAKAPI_ATTRIBUTE_DBID]);
          }
          if ($return['response']['attributes'][self::$KEYCLOAKAPI_ATTRIBUTE_MIOID] != $p_iKundeId) {
            //Not in the same club, this should be treated as an error??
            //Or a user not invited???
            //there was an error
            return array('STATISTIC' => 'ERR-User-Distinct-Mioid', 'RETURN' => $return['response']['attributes'][self::$KEYCLOAKAPI_ATTRIBUTE_MIOID]);
          }
          if ($return['response']['attributes'][self::$KEYCLOAKAPI_ATTRIBUTE_ADRID] != $p_aUserdata['adrid']) {
            //Not the same user, this should be treated as an error??
            //Or a user not invited???
            //there was an error
            if ($p_aUserdata['search'] == 'username') {  // When searching by username, we must be precise
              return array('STATISTIC' => 'ERR-User-Distinct-Adrid', 'RETURN' => $return['response']['attributes'][self::$KEYCLOAKAPI_ATTRIBUTE_ADRID]);
            }
            else {  // Otherwise a duplicate email for another member is ok
              return array('STATISTIC' => 'INF-Process-User-Not-Invited', 'RETURN' => $return);
            }
          }
          //Same Club, Same User
          return array('STATISTIC' => 'INF-Process-User-Status', 'RETURN' => $return['response']);
        }
      }
      elseif (is_array($return) && isset($return['ok']) && $return['ok'] === config('constants.keycloakSync.KEYCLOAKAPI_ERROR')) {

        //error_log( "Keycloak error: " . print_r($return, true) );
        if( is_array($return['response']) ) {
          $testStatus = $return['response']["status"];
        }
        elseif( is_object($return['response']) ) {
          $testStatus = $return['response']->status;
        }
        else {
          $testStatus = 0;
        }

        switch ($testStatus) {
          case 200:
            if ($return['message'] === 'USER NOT FOUND') {
              if ($p_aUserdata['search'] == 'userid') {
                //This means, user has a keycloak user id stored, but we did not find it
                return array('STATISTIC' => 'ERR-Process-User-Not-Found-by-ID', 'RETURN' => $userid);
              }
              //This means, user not invited, then WE INVITE
              return array('STATISTIC' => 'INF-Process-User-Not-Invited', 'RETURN' => $return);
            }
            else {
              return array('STATISTIC' => 'ERR-Ignore-me', 'RETURN' => '');
            }
          case 404:
            if ($p_aUserdata['search'] == 'userid') {
              //This means, user has a keycloak user id stored, but we did not find it
              return array('STATISTIC' => 'ERR-Process-User-Not-Found-by-ID', 'RETURN' => $userid);
            }
            //This means, user not invited, then WE INVITE
            return array('STATISTIC' => 'INF-Process-User-Not-Invited', 'RETURN' => $return);
          default:
            if ($testStatus > 899 && $testStatus < 1000) {
              //Errors related to CURL
              return array('STATISTIC' => 'ERR-User-Getting-Info', 'RETURN' => json_decode($return['response']['response'])->{'error'});
            }
            //Other type of error related to
            //the response
            return array('STATISTIC' => 'ERR-User-Getting-Info', 'RETURN' => json_encode($return['response']));
        }
        //We should not reach this place!!!
        return array('STATISTIC' => 'ERR-Ignore-me', 'RETURN' => '');
      }
      elseif (is_string($return)) {

        switch ($return) {
          case 'USER NOT FOUND':
            if ($p_aUserdata['search'] == 'userid') {
              //This means, user has a keycloak user id stored, but we did not find it
              return array('STATISTIC' => 'ERR-Process-User-Not-Found-by-ID', 'RETURN' => $userid);
            }
            //This means, user not invited, then WE INVITE
            return array('STATISTIC' => 'INF-Process-User-Not-Invited', 'RETURN' => $return);
          default:
            //Other type of error related to
            //the response
            return array('STATISTIC' => 'ERR-User-Getting-Info', 'RETURN' => $return);
        }
        //We shoud not reach this place!!!
        return array('STATISTIC' => 'ERR-Ignore-me', 'RETURN' => '');
      }
      //the options here are
      //Integer -3, missing username, which is not the case
      //Boolean FALSE, there was an error in the API
      return array('STATISTIC' => 'ERR-Keycloak-API', 'RETURN' => $return);
    }

    public static function setTime($time) {
        self::$KEYCLOAKAPI_TIME = $time;
    }
    public static function getRealmRoleName_CLUB_ADMINS(){
        return self::$KEYCLOAKAPI_REALM_ROLE_CLUB_ADMINS;
    }
    public static function getRealmRoleName_CLUB_MEMBERS(){
        return self::$KEYCLOAKAPI_REALM_ROLE_CLUB_MEMBERS;
    }
    public static function getRealmRoleName_CLUB_GUESTS(){
        return self::$KEYCLOAKAPI_REALM_ROLE_CLUB_GUESTS;
    }
    public static function getClientRoleName_ADMIN(){
        return self::$KEYCLOAKAPI_CLIENT_ROLE_ADMIN;
    }
    public static function getClientRoleName_USER(){
        return self::$KEYCLOAKAPI_CLIENT_ROLE_USER;
    }
    public static function getClientRoleName_GUEST(){
        return self::$KEYCLOAKAPI_CLIENT_ROLE_GUEST;
    }
    public static function getClientRoleName_FUNCTIONARY(){
        return self::$KEYCLOAKAPI_CLIENT_ROLE_FUNCTIONARY;
    }
    public static function getClientRoleName_SECRETARY(){
        return self::$KEYCLOAKAPI_CLIENT_ROLE_SECRETARY;
    }
    public static function getClientRoleName_EDITOR(){
        return self::$KEYCLOAKAPI_CLIENT_ROLE_EDITOR;
    }
    public static function getClientRoleName_MEMBER_LIGHT(){
        return self::$KEYCLOAKAPI_CLIENT_ROLE_MEMBER_LIGHT;
    }
    public static function getClientRoleName_ADMIN_LIGHT(){
        return self::$KEYCLOAKAPI_CLIENT_ROLE_ADMIN_LIGHT;
    }
    public static function getRealmRoleDescription_CLUB_ADMINS(){
        return self::$KEYCLOAKAPI_DESC_REALM_ROLE_CLUB_ADMINS;
    }
    public static function getRealmRoleDescription_CLUB_MEMBERS(){
        return self::$KEYCLOAKAPI_DESC_REALM_ROLE_CLUB_MEMBERS;
    }
    public static function getRealmRoleDescription_CLUB_GUESTS(){
        return self::$KEYCLOAKAPI_DESC_REALM_ROLE_CLUB_GUESTS;
    }
    public static function getClientRoleDescription_ADMIN(){
        return self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_ADMIN;
    }
    public static function getClientRoleDescription_USER(){
        return self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_USER;
    }
    public static function getClientRoleDescription_GUEST(){
        return self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_GUEST;
    }
    public static function getClientRoleDescription_FUNCTIONARY(){
        return self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_FUNCTIONARY;
    }
    public static function getClientRoleDescription_SECRETARY(){
        return self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_SECRETARY;
    }
    public static function getClientRoleDescription_EDITOR(){
        return self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_EDITOR;
    }
    public static function getClientRoleDescription_MEMBER_LIGHT(){
        return self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_MEMBER_LIGHT;
    }
    public static function getClientRoleDescription_ADMIN_LIGHT(){
        return self::$KEYCLOAKAPI_DESC_CLIENT_ROLE_ADMIN_LIGHT;
    }
    public static function getAttributeName_MIOID(){
        return self::$KEYCLOAKAPI_ATTRIBUTE_MIOID;
    }
    public static function getAttributeName_ADRID(){
        return self::$KEYCLOAKAPI_ATTRIBUTE_ADRID;
    }
    public static function getAttributeName_PHONE(){
        return self::$KEYCLOAKAPI_ATTRIBUTE_PHONE;
    }
    public static function getAttributeName_UPDDA(){
        return self::$KEYCLOAKAPI_ATTRIBUTE_UPDDA;
    }
    public static function getCLIENT_ID(){
        return self::$KEYCLOAKAPI_USER_CLIENT_ID;
    }

    /**
     * @param string $mode
     * @return void
     * @throws KeycloakException
     */
    public function synchronize( string $mode )
    {
        $dbdata = Databases::all();

        if( empty($dbdata) ) {
            throw new KeycloakException( "No databases found");
        }

        foreach( $dbdata as $aDbInfo ) {
            $vvClient = new DirectAccess($aDbInfo);

            switch ($mode) {
                case config('constants.keycloakSync.MODE_HOURLY'):
                    $vvClient->synchronizeHourly($aDbInfo["databaseid"]);
                    break;
                case config('constants.keycloakSync.MODE_DAILY'):
                    $vvClient->synchronizeDaily();
                    break;
                default:
                    throw new KeycloakException("Unknown mode: $mode");
            }

            $vvClient->vvClose();
            unset($vvClient);
        }
    }

    /**
     * Does a reverse sync, where the data from Keycloak is synchronized back to the database.
     * This is for manual use in special cases.
     *
     * @return void
     */
    public function synchronizeReverse()
    {
        $iInterval = 1;

        $dbdata = Databases::all();

        if( empty($dbdata) ) {
            throw new KeycloakException( "No databases found");
        }

        if (!self::setConfiguration(config('keycloak_api.keyCloakConfig'))['ok']) {
            return -1;
        }

        self::setTime((new DateTime('now'))->format('Y.m.d.H.i.s'));

        // Count users
        $userCnt = self::keycloakUsersCount();
        Log::debug("Total Users: $userCnt");

        foreach( $dbdata as $aDbInfo ) {
            Log::debug( "Working on DB " . $aDbInfo["databaseid"] );

            $vvClient = new DirectAccess($aDbInfo);

            $iCnt = 0;
            while ($iCnt <= $userCnt) {
                $iCnt += $iInterval;
                // Get all users from Keycloak
                $return = self::statusUser(array("first" => $iCnt, "max" => $iInterval, "q" => "vv_database_id:" . $aDbInfo["databaseid"]));
                Log::debug("User $iCnt\n" . print_r($return, TRUE) . "\n");
                if( $return["ok"] !== TRUE ) {
                    Log::debug("Error getting users from Keycloak, reached DB end");
                    break;
                }

                if( empty($return["response"]["attributes"]["vv_club_id"]) ||
                    empty($return["response"]["attributes"]["vv_member_id"]) ) {
                    Log::debug("Skipped user {$return["response"]["username"]} without vv_club_id or vv_member_id");
                    continue;
                }

                $aUser = array();
                //$aUser["dbid"] = $return["response"]["attributes"]["vv_database_id"];
                $aUser["mioid"] = $return["response"]["attributes"]["vv_club_id"];
                $aUser["adrid"] = $return["response"]["attributes"]["vv_member_id"];
                $aUser["date"] = date("d.m.Y H:i:s", $return["response"]["createdTimestamp"]/1000);
                $aUser["email"] = $return["response"]["email"];
                $aUser["username"] = $return["response"]["username"];
                $aUser["userid"] = $return["response"]["id"];
                $enabled = ($return["response"]["enabled"] ? TRUE : FALSE);
                $emailVerified = ($return["response"]["emailVerified"] ? TRUE : FALSE);
                if ($enabled && !$emailVerified) {
                    $status = config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED');
                }
                elseif (!$enabled && !$emailVerified) {
                    $status = config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILNOTVERIFIED');
                }
                elseif (!$enabled && $emailVerified) {
                    $status = config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILVERIFIED');
                }
                elseif ($enabled && $emailVerified) {
                    $status = config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILVERIFIED');
                }
                $aUser["status"] = $status;

                $aUser["role"] = self::keycloakUserGetClientRole(array("id" => $aUser["userid"]));
                if( is_array($aUser["role"]) ) {
                    $aUser["role"] = null;  // User doesn't have role
                }

                Log::debug("User: " . print_r($aUser, TRUE));

                if( $vvClient->vvSaveKCData($aUser["mioid"], $aUser["adrid"], array(
                    'userid' => $aUser["userid"],
                    'email' => $aUser["email"],
                    'user' => $aUser["username"],
                    'state' => $aUser["status"],
                    'role' => $aUser["role"],
                    'date' => $aUser["date"]
                )) ) {
                    Log::debug("User saved" );
                }
                else {
                    Log::debug("User not saved" );
                }
            }

            unset($vvClient);
        }
    }
}
