<?php
namespace App\Services;

use Illuminate\Support\Facades\Log;

/**
 * Class VVInterfaceBB
 * @package App\Services
 */
class VVInterfaceBB extends VVInterface
{
    const VV_EMAIL_PREFIX = "fibu-";
    const VV_EMAIL_SUFFIX = "@nomail.verein.cloud";
    const VV_EMAIL_SALT = "gy+0MrQewl";

    /**
     * VVInterfaceBB constructor.
     * @param $database_id
     * @param $p_iKundeId
     */
    public function __construct( $database_id = null, $p_iKundeId = null )
    {
        parent::__construct( $database_id, $p_iKundeId );
    }

    /**
     * Destructor that clears the data.
     */
    public function __destruct()
    {
        parent::__destruct();
    }

    /**
     * Generate pseudo email address from provided $db_id, $club_id and $user_id.
     * These are concatenated with _, then hashed and truncated to 10 characters.
     * Then enclosed in prefix and suffix, this giving a valid looking email address,
     * which can never be sent to.
     * This string is unique and repeatable for each user of the the VV.
     *
     * @param integer $db_id Database ID
     * @param integer $club_id Club ID
     * @param integer $user_id User ID
     * @return string Generated pseudo email address
     */
    private function generateUniqueEmail( $db_id, $club_id, $user_id )
    {
        $strEmailVariable = mb_substr( md5($db_id . "_" . $club_id . "_" . $user_id . self::VV_EMAIL_SALT), 0, 10 ) . random_int(1,9999999);
        return self::VV_EMAIL_PREFIX . $strEmailVariable . self::VV_EMAIL_SUFFIX;
    }

    /**
     * @param $club_id
     * @param $first
     * @param $max
     * @param $search
     * @return array|false
     */
    public function vvFindUser($club_id, $first, $max, $search)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;
        $this->loadGlobals();

        $iKundeId = $club_id;

        $aReturn = array();
        $aBenutzerData = FindBenutzer($conn, $club_id, $first, $max, $search);

        if (empty($aBenutzerData)) {
            return FALSE;
        }

        return $aBenutzerData;
    }

    /**
     * @param $club_id
     * @return int
     */
    public function vvGetUserCount($club_id)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;
        $this->loadGlobals();

        $iCount = CountBenutzer($conn, $club_id);

        return $iCount;
    }

    /**
     * @param integer $club_id
     * @param array $aCredData
     * @return void
     */
    public function vvSaveBBCreds($club_id, $aCredData)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;
        $this->loadGlobals();
        $mainApp->loadLib('/mio/lib/si_bbutler.php');

        // Add initialization of memberInitDone, depending on fibu type
        $iFibuType = BBGetFibuTypeAssistent( $conn, $club_id );
        if( !empty($iFibuType) && $iFibuType == BHB_FIBU_TYPE_EUR ) {
            $aCredData["memberinitdone"] = date("Y-m-d" );
        }
        else {
            $aCredData["memberinitdone"] = "";
        }

        return SaveBButlerCreds( $conn, $club_id, $aCredData );
    }

    /**
     * @param $club_id
     * @param $username
     * @param $user_id
     * @param $vid
     * @return array|bool
     */
    public function vvGetUser($db_id, $club_id, $username, $user_id, $vid)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;
        $this->loadGlobals();

        $iKundeId = $club_id;

        $aReturn = array();
        if (!empty($username)) {
            $aBenutzerData = GetUserByUsername($conn, $club_id, $username);
        }
        else {
            $aBenutzerData = GetUserData($conn, $user_id, $club_id);
        }

        if (empty($aBenutzerData)) {
            return FALSE;
        }

        if( $aBenutzerData["MioId"] != $club_id ) {
            return FALSE;
        }

        $aReturn["userId"] = $db_id."_".$club_id."_".$aBenutzerData["BenutzerId"];
        $aReturn["bbUserId"] = $aBenutzerData["BBUserId"];
        $aReturn["username"] = $aBenutzerData["UserName"];
        $aReturn["firstName"] = $aBenutzerData["Vorname"];
        $aReturn["lastName"] = $aBenutzerData["Nachname"];
        $aReturn["email"] = $this->generateUniqueEmail( $db_id, $club_id, $aBenutzerData["BenutzerId"] );
        if( empty($aBenutzerData["Mobil"]) ) {
            $aReturn["phone"] = "(none)";
        }
        else {
            $aReturn["phone"] = $aBenutzerData["Mobil"];
        }

        $aReturn["role"] = $aBenutzerData["Rolle"];
        $aReturn["vid"] = $vid;
        $aReturn["companyName"] = $aBenutzerData["ClubName"];
        //$aReturn["companyType"] = "e.V.";
        $aReturn["bbCustomerId"] = $aBenutzerData["BBCustId"];

        return $aReturn;
    }


    /**
     * @param $db_id
     * @param $club_id
     * @param $user_id
     * @param $pwhash
     * @return int HTTP return code
     */
    public function vvValidateUser($club_id, $user_id, $pwhash)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;
        $this->loadGlobals();
        $mainApp->loadLibs([
            '/lib/si_information.php',
            '/mio/lib/si_bbutler.php',
            '/mio/lib/api/crm_api_clubs.php'
        ]);
        $mainApp->loadClasses( array(
            '/classes/runtimelimit/runtimelimit.php',
        ));

        $iKundeId = $club_id;

        if (!empty($user_id) && !empty($pwhash)) {
            // Get user names of **this** Kunde to make sure this user is allowed to enter here.
            $aBenutzerData = GetUserData($conn, $user_id, $club_id);

            if (empty($aBenutzerData) || $aBenutzerData["Aktiv"] == 0 ) {
                // User doesn't exist or is inactive
                Log::debug( "ValidateUser($club_id, $user_id) failed: User doesn't exist or is inactive" );
                return 404;
            }

            // Check if club (still) is a FIBU user
            if( !ProductVariantHas($aBenutzerData["ProductVariant"], VV_PRODUCT_FIBU_PREMIUM) ) {
                Log::debug( "ValidateUser($club_id, $user_id) failed: FIBU not enabled" );
                return 401;
            }

            // check FIBU runtime limit
            $oRuntimeLimit = new \RuntimeLimit($conn, $iKundeId);
            if ($oRuntimeLimit->CheckLimit(OPHINEX_INFO_TYPE_FIBU_LAUFZEIT_ENDE) == -2) {
                Log::debug( "FIBU runtime limit expired, Kd $iKundeId, access denied" );
                /*
                 * Removed code to make an expired FIBU still reactivateable
                 *
                // Remove FIBU support for the club
                $input = array( "attributes" => array( "fibuPremium" => false ) );
                $fibuLight = null;
                $iNewProdVar = DefineNewProduktVariante($input, $aBenutzerData["ProductVariant"], $fibuLight);
                $aClubData = array();
                $aClubData["ProductVariant"] = $iNewProdVar;
                if( !UpdateKunde( $conn, $iKundeId, $aClubData ) ) {
                  Log::debug( "Error on update club $iKundeId, prod $iNewProdVar" );
                  return 500;
                }
                // Delete runtime limit
                if( !DeleteInformationByType($conn, $iKundeId, OPHINEX_INFO_TYPE_FIBU_LAUFZEIT_ENDE) ) {
                  Log::debug( "Error on DeleteInformationByType $iKundeId, prod $iNewProdVar" );
                  return 500;
                }
                // Delete Assistant info
                if( !BBDeleteAssistent( $conn, $iKundeId ) ) {
                  Log::debug( "Error on BBDeleteAssistent $iKundeId, prod $iNewProdVar" );
                  return 500;
                }
                // Delete BB Credentials
                if( !BBDeleteCredentials( $conn, $iKundeId ) ) {
                  Log::debug( "Error on BBDeleteCredentials $iKundeId, prod $iNewProdVar" );
                  return 500;
                }
                // Clean up user
                if( !BBCleanupBenutzer( $conn, $iKundeId ) ) {
                  Log::debug( "Error on BBCleanupBenutzer $iKundeId, prod $iNewProdVar" );
                  return 500;
                }
                */
                return 401;
            }

            $aAccess = GetUserBerchtigungen( $conn, $user_id , $club_id );
            if ($aAccess["prop" . TYP_MENU_BUCHHALTUNG] < TYP_BERECHT_Read ||
               empty($aAccess["prop" . TYP_MENU_FIBU_PREMIUM]) || $aAccess["prop" . TYP_MENU_FIBU_PREMIUM] < TYP_BERECHT_RdWrtDel ) {
                Log::debug( "ValidateUser($club_id, $user_id) failed: Missing permissions" );
                return 403;
            }

            //Log::error( "Hash compare $pwhash == " . $aBenutzerData["Password"] );
            if (strtolower($pwhash) != strtolower($aBenutzerData["Password"]) ) {
                Log::debug( "ValidateUser($club_id, $user_id) failed: Wrong password" );
                return 404;
            }
        }

        return 200;
    }
}
